from imnn.lfi import ApproximateBayesianComputation
import jax
import jax.numpy as np
import tensorflow_probability
from functools import partial
tfp = tensorflow_probability.substrates.jax
tfd = tfp.distributions


class PopulationMonteCarlo(ApproximateBayesianComputation):
    """ Population driven approximate Bayesian computation sample retrieval

    SEEMS TO BE A BUG IN THIS MODULE CURRENTLY!

    Approximate Bayesian computation tends to be very expensive because it
    depends on drawing samples from all over the prior and a lot of the domain
    may be extremely unlikely without good knowledge of the sampling
    distribution. To make sampling more efficient, the population Monte Carlo
    iteratively updates the proposal distribution for new samples based on the
    density of the closest summaries of simulations to the summary of the
    target.

    Parameters
    ----------
    acceptance_reached : float(n_targets,)
        The number of draws accepted over the number over the number of draws
        made for each target
    iterations : int(n_targets,)
        The number of iterations of population updates for each target
    total_draws : int(n_targets,)
        The total number of draws made (equivalent to number of simulations
        made) for each target
    """
    def __init__(self, target_data, prior, simulator, compressor, F=None,
                 gridsize=100, distance_measure=None):
        """ Constructor method

        Parameters
        ----------
        target_data: float(n_targets, input_data)
            Data (or batch of data) to infer parameter values at
        prior: fn
            A prior distribution which can be evaluated and sampled from
            (should also contain a ``low`` and a ``high`` attribute with
            appropriate ranges)
        simulator: fn
            A function which takes in a batch of parameter values (and random
            seeds) and returns a simulation made at each parameter value
        compressor: fn
            A function which takes a batch of simulations and returns their
            compressed summaries for each simulation (can be identity function
            for no compression)
        F : list or float({n_targets}, n_params, n_params) or None
            The Fisher information matrix to rescale parameter directions for
            measuring the distance. If None, this is set to the identity matrix
        gridsize : int or list, default=100
            The number of grid points to evaluate the marginal distribution on
            for every parameter (int) or each parameter (list)
        distance_measure : fn or None
            A distance measuring function between a batch of summaries and a
            single set of summaries for a target. If None, ``F_distance``
            (Euclidean scaled by the Fisher information) is used. This is only
        """
        super().__init__(
            target_data=target_data,
            prior=prior,
            simulator=simulator,
            compressor=compressor,
            gridsize=gridsize,
            F=F,
            distance_measure=distance_measure)
        self.acceptance_reached = np.zeros(self.n_targets)
        self.iterations = np.zeros(self.n_targets, dtype=np.int32)
        self.total_draws = np.zeros(self.n_targets, dtype=np.int32)

    def __call__(self, rng, n_points,
                 percentile=None, acceptance_ratio=0.1, max_iteration=10,
                 max_acceptance=1, max_samples=int(1e5),
                 n_initial_points=None, n_parallel_simulations=None,
                 proposed=None, summaries=None, distances=None, smoothing=None,
                 replace=False):
        """Run the PMC

        Parameters
        ----------
        rng : int(2,)
            A random number generator for generating simulations and randomly
            drawing parameter values from the prior
        n_points : int
            Number of points desired in the final (approximate) posterior
            (note that if `acceptance_ratio` is large then this approximation
            will be bad)
        percentile : int or None, default=None
            The percentage of points to define as the population, i.e. 75 for
            75% of the points. If None only the furthest point is moved one at
            a time
        acceptance_ratio : float
            The ratio of the number of accepted points to total proposals. When
            this gets small it suggests that points aren't being added to the
            population any more because the population is stationary
        max_iteration : int, default=10
            The cutoff number of iterations to break out of the while loop even
            if `acceptance_ratio` is not reached
        max_acceptance : int, default=1
            The cutoff number of attempts in a single iteration to attempt to
            get a sample accepted
        max_samples : int, default=100000
            The number of attempts to get parameter values from the truncated
            normal distribution
        n_initial_points : int or None, default=None
            The number of points to run in the initial ABC to define the
            population before the PMC starts. The PMC will always run from
            scratch if n_initial_points is passed
        n_parallel_simulations : int or None, default=None
            number of simulations to do at once, the innermost (closest)
            summaries are the only ones accepted, but this can massively reduce
            sampling time as long as the simulations can be easily parallelised
        proposed : float(any, n_params) or None, default=None
            A set of proposed parameters which have been used to premake
            simulations. Summaries of these simulations must also be passed as
            `summaries`. These can be used instead of running an initial ABC
            step
        summaries : float(any, n_summaries) or None, default=None
            A set of summaries of simulations which have been premade at
            parameter values corresponding to `proposed`. These can be used
            instead of running an initial ABC step
        distances : float(n_targets, any) or None, default=None
            An optional distance calculation from `summaries`, if this is not
            passed then distances is calculated in the call
        smoothing : float or None, default=None
            A Gaussian smoothing for the marginal distributions
        replace : bool, default=False
            Whether to replace the summaries, parameters and distances already
            obtained when running again

        Returns
        -------
        parameters container:
            All parameters with accepted and rejected attributes
        summaries container:
            All summaries with accepted and rejected attributes
        distances container:
            All distances with accepted and rejected attributes

        Raises
        ------
        ValueError
            if `n_initial_points` is less than `n_points`

        Todo
        ----
        type checking and pytests need writing
        """
        if n_initial_points is not None:
            if n_initial_points < n_points:
                raise ValueError(
                    "`n_initial_points` must be greater than or equal to " +
                    "the final number of points (`n_points`)")
            if n_parallel_simulations is not None:
                rng, *keys = jax.random.split(
                    rng, num=n_parallel_simulations + 1)
                proposed, summaries = jax.vmap(
                    lambda key: self.get_samples(
                        key, n_initial_points // n_parallel_simulations))(
                    np.array(keys))
            else:
                rng, *keys = jax.random.split(rng, num=n_initial_points + 1)
                proposed, summaries = jax.vmap(
                    lambda key: self.get_samples(key, 1))(np.array(keys))
            proposed = proposed.reshape((n_initial_points, -1))
            summaries = summaries.reshape((n_initial_points, -1))
            distances = jax.vmap(
                lambda target, F: self.distance_measure(
                    summaries, target, F))(self.target_summaries, self.F)
        elif (proposed is not None) and (summaries is not None):
            if distances is None:
                distances = jax.vmap(
                    lambda target, F: self.distance_measure(
                        summaries, target, F))(self.target_summaries, self.F)
        elif (self.parameters.all is not None) and (not replace):
            proposed = self.parameters.all.reshape((-1, self.n_params))
            summaries = self.summaries.all.reshape((-1, self.n_summaries))
            distances = jax.vmap(
                lambda target, F: self.distance_measure(
                    summaries, target, F))(self.target_summaries, self.F)
        else:
            raise ValueError(
                "`proposed` and `summaries` (and optionally `distances`) or " +
                "`n_initial_points` must be provided if PMC has not been " +
                "previously called")

        sample_indices = np.argsort(distances, axis=1)[:, :n_points]
        samples = jax.vmap(lambda x: proposed[x])(sample_indices)
        summaries = jax.vmap(lambda x: summaries[x])(sample_indices)
        distances = np.take(distances, sample_indices)

        weighting = self.prior.prob(samples)

        if percentile is None:
            ϵ_ind = -1
        else:
            ϵ_ind = int(percentile / 100 * n_points)

        key = np.array(jax.random.split(rng, num=self.n_targets))
        (rng, samples, summaries, distances, weighting, acceptance_reached,
         iteration_counter, total_draws) = jax.vmap(
            partial(
                self.move_samples,
                ϵ_ind=ϵ_ind,
                acceptance_ratio=acceptance_ratio,
                max_iteration=max_iteration,
                max_acceptance=max_acceptance,
                max_samples=max_samples,
                n_parallel_simulations=n_parallel_simulations))(
            key, samples, summaries, distances, weighting,
            self.target_summaries, self.F)
        self.set_samples(samples, summaries, distances=distances,
                         replace=replace)
        self.set_accepted(smoothing=smoothing)
        self.acceptance_reached = self.acceptance_reached + acceptance_reached
        self.iterations = self.iterations + iteration_counter
        self.total_draws = self.total_draws + total_draws
        print(
            f"Acceptance reached {self.acceptance_reached} in " +
            f"{self.iterations} iterations with a total of " +
            f"{self.total_draws} draws")
        return self.parameters, self.distances, self.summaries

    def move_samples(
            self, rng, samples, summaries, distances, weighting, target, F,
            ϵ_ind=None, acceptance_ratio=None, max_iteration=None,
            max_acceptance=None, max_samples=None,
            n_parallel_simulations=None):
        """ Loop which updates the population iteratively for individual target

        The values of the parameters corresponding to the most distant
        summaries to the summary of the target data are considered to be the
        means of a new proposal distribution with a covariance given by the
        population weighted samples (weighted from the prior in the first
        instance). From this proposal distribution a new sample (or a parallel
        set of samples) is proposed and a simulation made at each sample,
        compressed and the distance from the summary of the target data
        calculated. If this distance is within the population then that
        proposal is accepted and the weighting is updated with this new sample
        and the new furthest samples will be updated with this new proposal
        distribution on the next iteration. This is continued until the number
        of accepted draws is small compared to the total number of draws (this
        is the `acceptance_ratio`). At this point the distribution is somewhat
        stationary (as long as `acceptance_ratio` is small enough) and the
        samples approximately come from the target distribution.

        Parameters
        ----------
        rng : int(2,)
            A random number generator to draw the parameters and make
            simulations with
        samples : float(n_points, n_params)
            The parameter values from the ABC step
        summaries : float(n_points, n_summaries)
            The summaries of prerun simulations from the ABC step (at parameter
            values corresponding to the values in `samples`)
        distances : float(n_points,)
            The distances of the summaries from the summaries of a target
            from the ABC step
        weighting : float(n_points,)
            The value of the prior evaluated at the parameter values obtained
            in the ABC step
        target : float(n_summaries,)
            The summary of the target to infer the parameter values of
        F : float(n_params, n_params)
            The Fisher information matrix to rescale parameter directions for
            measuring the distance. This should be set to the identity matrix
            float(n_summaries, n_summaries) if the summaries are not parameter
            estimates
        ϵ_ind : float, default=None
            The index of the outer most population simulation
        acceptance_ratio : float, default=None
            The ratio of the number of accepted points to total proposals. When
            this gets small it suggests that points aren't being added to the
            population any more because the population is stationary
        max_iteration : int, default=None
            The cutoff number of iterations to break out of the while loop even
            if `acceptance_ratio` is not reached
        max_acceptance : int, default=None
            The cutoff number of attempts in a single iteration to attempt to
            get a sample accepted
        max_samples : int, default=None
            The number of attempts to get parameter values from the truncated
            normal distribution
        n_parallel_simulations : int or None, default=None
            number of simulations to do at once, the innermost (closest)
            summaries are the only ones accepted, but this can massively reduce
            sampling time as long as the simulations can be easily parallelised

        Returns
        -------
        int(2,):
            A random number generator to draw the parameters and make
            simulations with
        float(n_points, n_params):
            The accepted parameter values
        float(n_points, n_summaries):
            The summaries corresponding to the accepted parameter values, i.e.
            the closest summaries to the summary of the target data
        float(n_points,):
            The distances of the summaries from the summaries of the target
        float(n_points,):
            The value of the weighting from by the population
        float:
            The acceptance ratio reach by the final iteration
        int:
            The number of iterations run
        int:
            The total number of draws from the proposal distribution over all
            iterations

        Methods
        -------
        single_iteration_condition:
            Checks if the acceptance ratio or maximum iterations is reached
        single_iteration:
            Runs a single (or parallel set of) simulation(s), accepts if closer
        """
        def single_iteration_condition(args):
            """Checks if the acceptance ratio or maximum iterations is reached

            Parameters
            ----------
            args : tuple
                loop variables (described in `single_iteration`)

            Returns
            -------
            bool:
                True if acceptance_ratio is reached or the maximum number of
                iterations is reached
            """
            return np.logical_and(
                np.greater(args[-3], acceptance_ratio),
                np.less(args[-2], max_iteration))

        def single_iteration(args):
            """Runs single (parallel set of) simulation(s), accepts if closer

            Parameters
            ----------
            args : tuple
                - *rng* **int(2,)** -- A random number generator to draw the
                  parameters and make simulations with
                - *loc* **float(n_parallel_sims, n_params)** -- The current
                  iteration parameter values to use as the mean of the proposal
                  distribution for the next iterations parameters
                - *scale* **float(n_parallel_sims, n_params, n_params)** -- The
                  population weighted covariance for the proposal distribution
                - *summ* **float(n_parallel_sims, n_summaries)** -- The values
                  of the simulation summaries from this iteration
                - *dis* **float(n_parallel_sims,)** -- The distances between
                  the summaries being moved and the target summaries
                - *draws* **int** -- The number of draws taken so far
                - *accepted* **int** -- The number of proposals accepted so far
                - *acceptance_counter* **int** -- The total number of
                  iterations attempting to get an accepted proposal

            Returns
            -------
            tuple
                Same as input loop variables

            Methods
            -------
            single_acceptance_condition:
                checks if proposal has been accepted or max iterations reached
            single_acceptance:
                draws a proposal, simulates and compresses and checks distances
            """
            def single_acceptance_condition(args):
                """checks proposal has been accepted or max iterations reached

                Parameters
                ----------
                args : tuple
                    see loop variable in `single_iteration`

                Returns
                -------
                bool:
                    True if proposal not accepted and number of attempts to get
                    an accepted proposal not yet reached
                """
                return np.logical_and(
                    np.less(args[-2], 1),
                    np.less(args[-1], max_acceptance))

            def single_acceptance(args):
                """Draws a proposal, simulates and compresses, checks distance

                A new proposal is drawn from a truncated multivariate normal
                distribution whose mean is centred on the parameter to move and
                the covariance is set by the population. From this proposed
                parameter value a simulation is made and compressed and the
                distance from the target is calculated. If this distance is
                less than the current position then the proposal is accepted.

                Parameters
                ----------
                args : tuple
                    see loop variable in `single_iteration`

                Returns
                -------
                bool:
                    True if proposal not accepted and number of attempts to get
                    an accepted proposal not yet reached

                Todo
                ----
                Parallel sampling is currently commented out
                """
                (rng, loc, scale, summ, dis, draws, accepted,
                 acceptance_counter) = args
                rng, key = jax.random.split(rng)
                proposed, summaries = self.get_samples(
                    key, None, dist=tmvn(
                        loc, scale, self.prior.low, self.prior.high,
                        max_counter=max_samples))
                distances = np.squeeze(
                    self.distance_measure(
                        np.expand_dims(summaries, 0),
                        target,
                        F))
                # if n_parallel_simulations is not None:
                #     min_distance_index = np.argmin(distances)
                #     min_distance = distances[min_distance_index]
                #     closer = np.less(min_distance, ϵ)
                #     loc = jax.lax.cond(
                #         closer,
                #         lambda _ : proposed[min_distance_index],
                #         lambda _ : loc,
                #         None)
                #     summ = jax.lax.cond(
                #         closer,
                #         lambda _ : summaries[min_distance_index],
                #         lambda _ : summ,
                #         None)
                #     dis = jax.lax.cond(
                #         closer,
                #         lambda _ : distances[min_distance_index],
                #         lambda _ : dis,
                #         None)
                #     iteration_draws = n_parallel_simulations \
                #         - np.isinf(distances).sum()
                #     draws += iteration_draws
                #     accepted = closer.sum()
                # else:
                closer = np.less(distances, np.min(dis))
                loc = jax.lax.cond(
                    closer,
                    lambda _: proposed,
                    lambda _: loc,
                    None)
                summ = jax.lax.cond(
                    closer,
                    lambda _: summaries,
                    lambda _: summ,
                    None)
                dis = jax.lax.cond(
                    closer,
                    lambda _: distances,
                    lambda _: dis,
                    None)
                iteration_draws = 1 - np.isinf(distances).sum()
                draws += iteration_draws
                accepted = closer.sum()
                return (rng, loc, scale, summ, dis, draws, accepted,
                        acceptance_counter + 1)

            (rng, samples, summaries, distances, weighting, acceptance_reached,
             iteration_counter, total_draws) = args
            n_to_move = samples[ϵ_ind:].shape[0]
            cov = self.w_cov(samples, weighting)
            scale = np.linalg.cholesky(cov)
            rng, *keys = jax.random.split(rng, num=n_to_move + 1)

            results = jax.vmap(
                lambda key, loc, scale, summaries, distances, draws, accepted,
                acceptance_counter: jax.lax.while_loop(
                    single_acceptance_condition,
                    single_acceptance,
                    (key, loc, scale, summaries, distances, draws, accepted,
                     acceptance_counter)))(
                np.array(keys),
                samples[ϵ_ind:],
                np.repeat(np.expand_dims(scale, 0), n_to_move, axis=0),
                summaries[ϵ_ind:],
                distances[ϵ_ind:],
                np.zeros(n_to_move, dtype=np.int32),
                np.zeros(n_to_move, dtype=np.int32),
                np.zeros(n_to_move))

            weighting = jax.vmap(
                lambda proposed: (
                    self.prior.prob(proposed)
                    / (np.sum(weighting * tfd.MultivariateNormalTriL(
                        loc=proposed,
                        scale_tril=np.repeat(
                            np.expand_dims(scale, 0),
                            samples.shape[0],
                            axis=0)).prob(proposed)))))(
                np.vstack([samples[:ϵ_ind], results[1]]))
            samples = jax.ops.index_update(
                samples,
                jax.ops.index[ϵ_ind:, :],
                results[1])
            summaries = jax.ops.index_update(
                summaries,
                jax.ops.index[ϵ_ind:, :],
                results[3])
            distances = jax.ops.index_update(
                distances,
                jax.ops.index[ϵ_ind:],
                results[4])
            sample_indices = np.argsort(distances)
            samples = samples[sample_indices]
            summaries = summaries[sample_indices]
            distances = distances[sample_indices]
            weighting = weighting[sample_indices]
            acceptance_reached = results[-2].sum() / results[-3].sum()
            return (rng, samples, summaries, distances, weighting,
                    acceptance_reached, iteration_counter + 1,
                    total_draws + results[-3].sum())

        acceptance_reached = np.inf
        iteration_counter = 0
        total_draws = 0
        return jax.lax.while_loop(
            single_iteration_condition,
            single_iteration,
            (rng, samples, summaries, distances, weighting, acceptance_reached,
             iteration_counter, total_draws))

    def set_samples(self, parameters, summaries, distances=None,
                    replace=False):
        """Fills containers and calculates distances

        All parameters and summaries of simulations made at those parameters
        (and their distance from the summary of the observed data) are kept in
        containers for ease of use. This function sets the ``all`` attribute of
        these containers with the passed summaries and parameters (and
        distances if provided or it calculates the distances). These are
        concatenated to existing values unless ``replace = True`` in which the
        existing values are removed and overwritten with the passed summaries,
        parameters and distances.

        Parameters
        ----------
        parameters : float(n_targets, n_points, n_params)
            The pre-obtained parameter values to set to the class
        summaries : float(n_targets, n_points, n_summaries)
            The summaries of prerun simulations (at parameter values
            corresponding to the values in `parameters`)
        distances : float(n_targets, n_points) or None, default=None
            The distances of the summaries from the summaries of each target.
            If None then the distances will be calculated
        replace : bool, default=False
            Whether to replace the summaries, parameters and distances already
            obtained
        """
        if distances is None:
            distances = jax.vmap(
                lambda target, F: self.distance_measure(
                    summaries, target, F))(self.target_summaries, self.F)
        if (self.parameters.all is None) or (replace):
            self.parameters.all = parameters
            self.summaries.all = summaries
            self.distances.all = distances
        else:
            self.parameters.all = np.concatenate(
                [self.parameters.all, parameters], axis=1)
            self.summaries.all = np.concatenate(
                [self.summaries.all, summaries], axis=1)
            self.distances.all = np.concatenate(
                [self.distances.all, distances], axis=1)
        self.parameters.size = self.parameters.all.shape[0]
        self.summaries.size = self.summaries.all.shape[0]
        self.distances.size = self.distances.all.shape[-1]

    def set_accepted(self, smoothing=None):
        """Container values to list of accepted attributes, builds marginals

        Parameters
        ----------
        smoothing : float or None, default=None
            A Gaussian smoothing for the marginal distributions
        """
        self.parameters.accepted = [params for params in self.parameters.all]
        self.parameters.n_accepted = [
            params.shape[0] for params in self.parameters.all]
        self.summaries.accepted = [
            summaries for summaries in self.summaries.all]
        self.summaries.n_accepted = [
            summaries.shape[0] for summaries in self.summaries.all]
        self.distances.accepted = [
            distances for distances in self.distances.all]
        self.distances.n_accepted = [
            distances.shape[0] for distances in self.distances.all]
        self.parameters.rejected = None
        self.summaries.rejected = None
        self.distances.rejected = None
        self.marginals = self.get_marginals(smoothing=smoothing)

    def w_cov(self, proposed, weighting):

        weighted_samples = proposed * weighting[:, np.newaxis]
        return weighted_samples.T.dot(
            weighted_samples) / weighting.T.dot(weighting)


class tmvn():
    """Truncated multivariate normal distribution (for sampling)

    Parameters
    ----------
    loc : float(any, n_params)
        The mean of any number of input distributions
    scale : float(any, n_params, n_params)
        The cholesky matrix of the covariance
    low : float(any, n_params)
        The minimum value for the truncation for each parameter
    high : float(any, n_params)
        The maximum value for the truncation for each parameter
    n_samples : int or None
        The number of different distributions to make (equivalent to
        `batch_shape`)
    n_params : int
        The number of parameters (equivalent to `event_shape`)
    max_counter : int
        Number of iterations to try to get accepted samples (within truncation)
    """
    def __init__(self, loc, scale, low, high, max_counter=int(1e3)):
        self.loc = loc
        self.scale = scale
        self.low = low
        self.high = high
        if len(loc.shape) > 1:
            self.n_samples = loc.shape[0]
        else:
            self.n_samples = None
        self.n_params = low.shape[0]
        self.max_counter = max_counter

    def mvn(self, rng, loc):
        u = jax.random.normal(rng, shape=(self.n_params,))
        return loc + u.dot(self.scale)

    def w_cond(self, args):
        _, loc, counter = args
        return np.logical_and(
            np.logical_or(
                np.any(np.greater(loc, self.high)),
                np.any(np.less(loc, self.low))),
            np.less(counter, self.max_counter))

    def __sample(self, args):
        rng, loc, counter = args
        rng, key = jax.random.split(rng)
        return (rng, self.mvn(key, loc), counter + 1)

    def _sample(self, rng, loc):
        rng, key = jax.random.split(rng)
        _, loc, counter = jax.lax.while_loop(
            self.w_cond,
            self.__sample,
            (rng, self.mvn(key, loc), 0))
        return jax.lax.cond(
            np.greater_equal(counter, self.max_counter),
            lambda _: np.nan * np.ones((self.n_params,)),
            lambda _: loc,
            None)

    def _sample_n(self, rng, loc, n=None):
        if n is None:
            return self._sample(rng, loc)
        else:
            key = jax.random.split(rng, num=n)
            return jax.vmap(self._sample)(
                key, np.repeat(loc[np.newaxis], n, axis=0))

    def sample(self, shape=None, seed=None):
        if shape is None:
            if self.n_samples is None:
                return self._sample_n(seed, self.loc)
            else:
                key = jax.random.split(seed, num=self.n_samples)
                return jax.vmap(
                    lambda key, loc: self._sample_n(key, loc))(key, self.loc)
        elif len(shape) == 1:
            if self.n_samples is None:
                return self._sample_n(seed, self.loc, n=shape[0])
            else:
                key = jax.random.split(seed, num=self.n_samples)
                return jax.vmap(
                    lambda key, loc: self._sample_n(key, loc, n=shape[0]))(
                    key, self.loc)
        else:
            key = jax.random.split(seed, num=shape[-1])
            return jax.vmap(
                lambda key: self.sample(
                    shape=tuple(shape[:-1]), seed=key), out_axes=-2)(key)



import jax
import jax.numpy as np
from imnn.lfi import LikelihoodFreeInference
from imnn.utils import container
from scipy.ndimage import gaussian_filter
from imnn.utils.utils import _check_input


class ApproximateBayesianComputation(LikelihoodFreeInference):
    """Approximate Bayesian computation

    Approximate Bayesian computation involves drawing parameter values from a
    prior, :math:`\\boldsymbol{\\theta}_i\\leftarrow p(\\boldsymbol{\\theta})`,
    and generating simulations at this value using a model,
    :math:`{\\bf d}_i = f(\\boldsymbol{\\theta}_i)`, which can then be
    compressed down to some informative summary,
    :math:`{\\bf x}_i =g({\\bf d}_i)`. By compressing some observed target to
    get :math:`{\\bf x}^\\textrm{obs} =g({\\bf d}^\\textrm{obs})` we can
    simulate from the prior distribution of possible parameter values to create
    a set of summaries :math:`\\{{\\bf x}_i|i\\in[1, n_\\textrm{sims}]\\}`. The
    approximate posterior distribution is then obtained as the samples of
    parameter values whose summaries of simulations have an infinitely small
    difference from the summary of the target observation:

    .. math::

        p(\\boldsymbol{\\theta}|{\\bf d}^\\textrm{obs})\\approx
        \\{\\boldsymbol{\\theta}_i|
        i~\\textrm{where}~\\lim_{\\epsilon\\to0}
        D(g(f(\\boldsymbol{\\theta}_i), g({\\bf d}^\\textrm{obs}))<\\epsilon\\}

    Note that the distance measure :math:`D(x, y)` is somewhat ambiguous, but
    in the infinitessimal limit will be more or less flat. Sensible choices for
    this distance measure can be made, i.e. for parameter estimates a Fisher
    scaled squared distance, but it is not wholly trivial in all cases.

    This class contains several methods for making simulations, compressing
    them and doing distance based acceptance and rejection of the parameters
    used to make the simulations from the compressed summaries of target data.
    In particular either a single set of simulations can be run and the
    distance calculated, simulations can be run until a desired number of
    accepted points within achosen distance are obtained or if there are
    already a set of run simulations, different distance values can be chosen
    for acceptance and rejection of points. There are also several plotting
    methods for corner plots.

    Note that if the summaries of the data are not parameter estimates then the
    distance measure must be optimal for the summary space to make sense.

    Parameters
    ----------
    target_summaries : float(n_targets, n_summaries)
        The compressed data to be infered
    n_summaries : int
        The size of the output of the compressor for a single simulation
    n_targets : int
        The number of different targets to be infered
    F : float(n_params, n_params)
        The Fisher information matrix for rescaling distance measure
    parameters : utils container
        Holds accepted and rejected parameters and some summaries of results
    summaries : utils container
        Holds accepted and rejected summaries and some summaries of results
    distances : utils container
        Holds accepted and rejected distances of summaries to targets and
        some summaries of results

    Methods
    -------
    simulator
        A simulator which takes in an array of parameters and returns
        simulations made at those parameter values
    compressor
        A function which takes in an array of simulations and compresses them.
        If no compression is needed this can be an identity function
    distance_measure:
        Either ``F_distance`` or ``euclidean_distance`` depending on inputs
    """
    def __init__(self, target_data, prior, simulator, compressor,
                 gridsize=100, F=None, distance_measure=None):
        """Constructor method

        Parameters
        ----------
        target_data: float(n_targets, input_data)
            Data (or batch of data) to infer parameter values at
        prior: fn
            A prior distribution which can be evaluated and sampled from
            (should also contain a ``low`` and a ``high`` attribute with
            appropriate ranges)
        simulator: fn
            A function which takes in a batch of parameter values (and random
            seeds) and returns a simulation made at each parameter value
        compressor: fn
            A function which takes a batch of simulations and returns their
            compressed summaries for each simulation (can be identity function
            for no compression)
        gridsize : int or list, default=100
            The number of grid points to evaluate the marginal distribution on
            for every parameter (int) or each parameter (list)
        F : list or float({n_targets}, n_params, n_params) or None
            The Fisher information matrix to rescale parameter directions for
            measuring the distance. If None, this is set to the identity matrix
        distance_measure : fn or None
            A distance measuring function between a batch of summaries and a
            single set of summaries for a target. If None, ``F_distance``
            (Euclidean scaled by the Fisher information) is used. This is only
        """
        super().__init__(
            prior=prior,
            gridsize=gridsize)
        self.simulator = simulator
        self.compressor = compressor
        if len(target_data.shape) == 1:
            target_data = np.expand_dims(target_data, 0)
        self.target_summaries = self.compressor(target_data)
        self.n_summaries = self.target_summaries.shape[-1]
        self.n_targets = self.target_summaries.shape[0]
        if F is not None:
            if self.n_summaries != self.n_params:
                raise ValueError(
                    "If using the Fisher information to scale the distance " +
                    "then the compressor must return parameter estimates. " +
                    "The compressor returns summaries with shape " +
                    f"{(self.n_summaries,)}, but should return estamites " +
                    f"with shape {(self.n_params)} if the prior is correct.")
            if isinstance(F, list):
                self.F = np.stack([
                    _check_input(f, (self.n_params, self.n_params), "F")
                    for f in F], 0)
            else:
                if F.shape == (self.n_params, self.n_params):
                    self.F = np.expand_dims(F, 0)
                    if self.n_targets > 1:
                        self.F = np.repeat(self.F, self.n_targets)
                else:
                    self.F = _check_input(
                        F, (self.n_targets, self.n_params, self.n_params), "F")
            self.distance_measure = self.F_distance

        if distance_measure is not None:
            self.distance_measure = distance_measure
        elif F is None:
            self.distance_measure = self.euclidean_distance
            self.F = np.zeros(self.n_targets)

        self.parameters = container()
        self.parameters.n_params = self.n_params
        self.summaries = container()
        self.summaries.n_summaries = self.n_summaries
        self.distances = container()
        self.distances.n_targets = self.n_targets

    def __call__(self, ϵ=None, rng=None, n_samples=None, parameters=None,
                 summaries=None, min_accepted=None, max_iterations=10,
                 smoothing=None, replace=False):
        """Run the ABC

        Parameters
        ----------
        ϵ : float or float(n_targets) or None, default=None
            The acceptance distance between summaries from simulations and the
            summary of the target data. A different epsilon can be passed for
            each target. If None (and `rng` and `n_samples` are not None) then
            samples are obtained but no acceptance and rejection takes place
        rng : int(2,) or None, default=None
            A random number generator for generating simulations and randomly
            drawing parameter values from the prior. Can be None if passing
            `parameters` and `summaries` to set prerun simulations or if `ϵ`
            is not None for resetting the accepted parameters and summaries
        n_samples : int or None, default=None
            Number of simulations to make (or how many to make each iteration
            if `min_accepted` is not None). Can be None if passing `parameters`
            and `summaries` to set prerun simulations or if `ϵ` is not None for
            resetting the accepted parameters and summaries
        parameters : float(any, n_params) or None, default=None
            The pre-obtained parameter values to set to the class (`summaries`
            must not be None)
        summaries : float(any, n_summaries) or None, default=None
            The summaries of prerun simulations (at parameter values
            corresponding to the values in `parameters`) to set the class
            attributes
        min_accepted : int or None, default=None
            A minimum number of accepted samples desired within a chosen `ϵ`
            value. If None and `ϵ`, `rng`, `n_samples` are not None then just
            `n_samples` will be run
        max_iterations : int, default=10
            A maximum number of while loop iterations to run when getting a
            minimum number of accepted samples
        smoothing : float or None, default=None
            A Gaussian smoothing for the marginal distributions
        replace : bool, default=False
            Whether to replace the summaries, parameters and distances already
            obtained when running again

        Returns
        -------
        parameters container:
            All parameters with accepted and rejected attributes
        summaries container:
            All summaries with accepted and rejected attributes
        distances container:
            All distances with accepted and rejected attributes

        Todo
        ----
        type checking and pytests
        """
        if ((self.parameters.all is None)
                and ((rng is None) or (n_samples is None))
                and ((parameters is None) or (summaries is None))):
            raise ValueError(
                "No samples currently available. If running simulations " +
                "`rng` must be a JAX prng and `n_samples` an integer number " +
                "of samples. If using premade summaries `parameters` and " +
                "`summaries` must be numpy arrays.")
        if ((min_accepted is not None) and (ϵ is None)):
            raise ValueError("`ϵ` must be passed if passing `min_accepted`")
        if (rng is not None) and (n_samples is not None):
            if ((ϵ is not None) and (min_accepted is not None)):
                self.set_samples(
                    *self.get_min_accepted(
                        rng=rng, ϵ=ϵ, accepted=min_accepted,
                        n_simulations=n_samples,
                        max_iterations=max_iterations),
                    replace=True)
            else:
                self.set_samples(
                    *self.get_samples(rng=rng, n_samples=n_samples),
                    replace=replace)
        if (parameters is not None) and (summaries is not None):
            self.set_samples(parameters, summaries, replace=replace)
        if ϵ is not None:
            self.set_accepted(ϵ, smoothing=smoothing)
        return self.parameters, self.distances, self.summaries

    def set_samples(self, parameters, summaries, distances=None,
                    replace=False):
        """Fills containers and calculates distances

        All parameters and summaries of simulations made at those parameters
        (and their distance from the summary of the observed data) are kept in
        containers for ease of use. This function sets the ``all`` attribute of
        these containers with the passed summaries and parameters (and
        distances if provided or it calculates the distances). These are
        concatenated to existing values unless ``replace = True`` in which the
        existing values are removed and overwritten with the passed summaries,
        parameters and distances.

        Parameters
        ----------
        parameters : float(any, n_params)
            The pre-obtained parameter values to set to the class
        summaries : float(any, n_summaries)
            The summaries of prerun simulations (at parameter values
            corresponding to the values in `parameters`)
        distances : float(n_targets, any) or None, default=None
            The distances of the summaries from the summaries of each target.
            If None then the distances will be calculated
        replace : bool, default=False
            Whether to replace the summaries, parameters and distances already
            obtained
        """
        if distances is None:
            distances = jax.vmap(
                lambda target, F: self.distance_measure(
                    summaries, target, F))(self.target_summaries, self.F)
        if (self.parameters.all is None) or (replace):
            self.parameters.all = parameters
            self.summaries.all = summaries
            self.distances.all = distances
        else:
            self.parameters.all = np.concatenate(
                [self.parameters.all, parameters], axis=0)
            self.summaries.all = np.concatenate(
                [self.summaries.all, summaries], axis=0)
            self.distances.all = np.concatenate(
                [self.distances.all, distances], axis=0)
        self.parameters.size = self.parameters.all.shape[0]
        self.summaries.size = self.summaries.all.shape[0]
        self.distances.size = self.distances.all.shape[-1]

    def get_samples(self, rng, n_samples, dist=None):
        """Draws parameters from prior and makes and compresses simulations

        Simulations are done `n_samples` in parallel

        Parameters
        ----------
        rng : int(2,)
            A random number generator to draw the parameters and make
            simulations with
        n_samples : int
            Number of parallel samples to draw
        dist : fn or None, default=None
            A distribution to sample, if None the samples will be drawn from
            the prior set in the class initialisation
        """
        rng, key = jax.random.split(rng)
        if dist is None:
            parameters = self.prior.sample(n_samples, seed=rng)
        else:
            parameters = dist.sample(n_samples, seed=rng)
        summaries = self.compressor(self.simulator(key, parameters))
        return parameters, summaries

    def set_accepted(self, ϵ, smoothing=None):
        """Sets the accepted and rejected attributes of the containers

        Using a distance (or list of distances for each target) cutoff between
        simulation summaries and summaries from some target the accepted and
        rejected parameter values (and summaries) can be defined. These points
        are used to make an approximate set of marginal distributions for
        plotting based on histogramming the points - where smoothing can be
        performed on the histogram to avoid undersampling artefacts.

        Parameters
        ----------
        ϵ : float or float(n_targets)
            The acceptance distance between summaries from simulations and the
            summary of the target data. A different epsilon can be passed for
            each target.
        smoothing : float or None, default=None
            A Gaussian smoothing for the marginal distributions

        Methods
        -------
        get_accepted:
            Returns a boolean array with whether summaries are within `ϵ`
        """
        def get_accepted(distances, ϵ):
            """ Returns a boolean array with whether summaries are within `ϵ`

            Parameters
            ----------
            distances : float(any)
                The distances between the summary of a target and the summaries
                of the run simulations
            ϵ : float
                The acceptance distance between summaries from simulations and
                the summary of the target data
            """
            return np.less(distances, ϵ)

        if not isinstance(ϵ, list):
            accepted = jax.vmap(
                lambda distances: get_accepted(distances, ϵ))(
                self.distances.all)
        else:
            accepted = jax.vmap(get_accepted)(self.distances.all, ϵ)

        rejected = ~accepted
        accepted_inds = [np.argwhere(accept)[:, 0] for accept in accepted]
        rejected_inds = [np.argwhere(reject)[:, 0] for reject in rejected]
        self.parameters.ϵ = ϵ
        self.summaries.ϵ = ϵ
        self.distances.ϵ = ϵ
        self.parameters.accepted = [self.parameters.all[inds]
                                    for inds in accepted_inds]
        self.parameters.n_accepted = np.array([
            self.parameters.accepted[i].shape[0]
            for i in range(self.n_targets)])
        self.summaries.accepted = [self.summaries.all[inds]
                                   for inds in accepted_inds]
        self.summaries.n_accepted = np.array([
            self.summaries.accepted[i].shape[0]
            for i in range(self.n_targets)])
        self.distances.accepted = [
            self.distances.all[i, inds]
            for i, inds in enumerate(accepted_inds)]
        self.distances.n_accepted = np.array([
            self.distances.accepted[i].shape[0]
            for i in range(self.n_targets)])
        self.parameters.rejected = [
            self.parameters.all[inds] for inds in rejected_inds]
        self.summaries.rejected = [
            self.summaries.all[inds] for inds in rejected_inds]
        self.distances.rejected = [
            self.distances.all[i, inds]
            for i, inds in enumerate(rejected_inds)]
        self.marginals = self.get_marginals(smoothing=smoothing)

    def get_min_accepted(self, rng, ϵ, accepted, n_simulations=1,
                         max_iterations=10, smoothing=None, verbose=True):
        """Iteratively run ABC until a minimum number of samples are accepted

        Setting a maximum distance that simulation summaries can be allowed to
        be from the summaries of each target an iterative scheme can be used to
        get ``n_samples`` simulations at a time, compress them and calculate
        the distances until the minimum number of desired samples are within
        the allowed cutoff distance. If multiple targets are being inferred
        then the simulations are run until all targets have at least the
        minimum number of accepted samples.

        Parameters
        ----------
        rng : int(2,)
            A random number generator to draw the parameters and make
            simulations with
        ϵ : float or float(n_targets)
            The acceptance distance between summaries from simulations and the
            summary of the target data. A different epsilon can be passed for
            each target.
        accepted : int
            The minimum number of samples to be accepted within the `ϵ` cutoff
        n_simulations : int, default=1
            The number of simulations to do at once
        max_iterations : int, default=10
            Maximum number of iterations in the while loop to prevent infinite
            runs. Note if max_iterations is reached then there are probably
            not the required number of accepted samples
        smoothing : float or None, default=None
            A Gaussian smoothing for the marginal distributions
        verbose : bool, default=True
            Whether to print out the running stats (number accepted,
            iterations, etc)

        Returns
        -------
        float(any, n_params)
            All parameters values drawn
        float(any, n_summaries)
            All summaries of simulations made
        float(n_target, any)
            The distance of every summary to each target
        """
        @jax.jit
        def loop(inputs):
            return jax.lax.while_loop(loop_cond, loop_body, inputs)

        def loop_cond(inputs):
            return np.logical_and(
                np.less(np.min(inputs[-2]), accepted),
                np.less(inputs[-1], max_iterations))

        def loop_body(inputs):
            rng, parameters, summaries, distances, n_accepted, iteration = \
                inputs
            rng, key = jax.random.split(rng)
            parameter_samples = self.prior.sample(n_simulations, seed=key)
            rng, key = jax.random.split(rng)
            summary_samples = self.compressor(
                self.simulator(key, parameter_samples))
            distance_samples = jax.vmap(
                lambda target, F: self.distance_measure(
                    summary_samples, target, F))(
                self.target_summaries, self.F)
            indices = jax.lax.dynamic_slice(
                np.arange(n_simulations * max_iterations),
                [n_simulations * iteration],
                [n_simulations])
            parameters = jax.ops.index_update(
                parameters,
                jax.ops.index[indices],
                parameter_samples)
            summaries = jax.ops.index_update(
                summaries,
                jax.ops.index[indices],
                summary_samples)
            distances = jax.ops.index_update(
                distances,
                jax.ops.index[:, indices],
                distance_samples)
            n_accepted = np.int32(np.less(distances, ϵ).sum(1))
            return rng, parameters, summaries, distances, n_accepted, \
                iteration + np.int32(1)

        parameters = np.ones((max_iterations * n_simulations, self.n_params))
        summaries = np.ones(
            (max_iterations * n_simulations,
             self.n_summaries))
        distances = np.ones((
            self.n_targets,
            max_iterations * n_simulations)) * np.inf
        if self.parameters.all is not None:
            parameters = np.vstack([parameters, self.parameters.all])
            summaries = np.vstack([summaries, self.summaries.all])
            distances = np.hstack([distances, self.distances.all])
            if self.parameters.n_accepted is None:
                self.set_accepted(ϵ, smoothing=smoothing)
            n_accepted = np.int32(self.parameters.n_accepted)
        else:
            n_accepted = np.zeros(self.n_targets, dtype=np.int32)
        current_accepted = n_accepted
        iteration = 0
        _, parameters, summaries, distances, n_accepted, iteration = loop(
            (rng, parameters, summaries, distances, n_accepted, iteration))
        keep = ~np.any(np.isinf(distances), 0)
        parameters = parameters[keep]
        summaries = summaries[keep]
        distances = distances[:, keep]
        if verbose:
            print(f"{n_accepted - current_accepted} accepted in last ",
                  f"{iteration} iterations ",
                  f"({n_simulations * iteration} simulations done).")
        return parameters, summaries, distances

    def set_marginals(self, accepted_parameters=None, ranges=None,
                      gridsize=None, smoothing=None):
        """Wrapper for ``get_marginals`` to set the ``marginals`` attribute

        Parameters
        ----------
        accepted_parameters : float(any, n_params) or None, default=None
            An array of all accepted parameter values
        ranges : list or None, default=None
            A list of arrays containing the bin centres for the marginal
            distribution obtained by histogramming for each parameter.
        gridsize : list or None, default=None
            The number of grid points to evaluate the marginal distribution on
            for each parameter
        smoothing : float or None, default=None
            A Gaussian smoothing for the marginal distributions
        """
        self.marginals = self.get_marginals(
            accepted_parameters=accepted_parameters, ranges=ranges,
            gridsize=gridsize, smoothing=smoothing)

    def get_marginals(self, accepted_parameters=None, ranges=None,
                      gridsize=None, smoothing=None):
        """ Creates the 1D and 2D marginal distribution list for plotting

        Using list of parameter values (accepted by the ABC) an approximate set
        of marginal distributions for plotting are created based on
        histogramming the points. Smoothing can be performed on the histogram
        to avoid undersampling artefacts.

        For every parameter the full distribution is summed over every other
        parameter to get the 1D marginals and for every combination the 2D
        marginals are calculated by summing over the remaining parameters. The
        list is made up of a list of n_params lists which contain n_columns
        number of objects.

        Parameters
        ----------
        accepted_parameters : float(any, n_params) or None, default=None
            An array of all accepted parameter values. If None, the accepted
            parameters from the `parameters` class attribute are used
        ranges : list or None, default=None
            A list of arrays containing the bin centres for the marginal
            distribution obtained by histogramming for each parameter. If None
            the ranges are constructed from the ranges of the prior
            distribution.
        gridsize : list or None, default=None
            The number of grid points to evaluate the marginal distribution on
            for each parameter. This needs to be set if ranges is passed (and
            different from the gridsize set on initialisation)
        smoothing : float or None, default=None
            A Gaussian smoothing for the marginal distributions. Smoothing not
            done if smoothing is None

        Returns
        -------
        list of lists:
            The 1D and 2D marginal distributions for each parameter (of pair)
        """
        if accepted_parameters is None:
            accepted_parameters = self.parameters.accepted
        if ranges is None:
            ranges = [
                np.hstack([range, np.array([range[1] - range[0]])])
                - (range[1] - range[0]) / 2 for range in self.ranges]
        if gridsize is None:
            gridsize = self.gridsize
        if smoothing is not None:
            def smooth(x):
                return gaussian_filter(x, smoothing, mode="mirror")
        else:
            def smooth(x):
                return x
        marginals = []
        for row in range(self.n_params):
            marginals.append([])
            for column in range(self.n_params):
                if column == row:
                    marginals[row].append(
                        np.array([
                            smooth(np.histogram(
                                parameters[:, column],
                                bins=ranges[column],
                                density=True)[0])
                            for parameters in accepted_parameters]))
                elif column < row:
                    marginals[row].append(
                        np.array([
                            smooth(np.histogramdd(
                                parameters[:, [column, row]],
                                bins=[ranges[column], ranges[row]],
                                density=True)[0])
                            for parameters in accepted_parameters]))
        return marginals

    def scatter_plot(self, ax=None, ranges=None, points=None, label=None,
                     axis_labels=None, colours=None, hist=True, s=5, alpha=1.,
                     figsize=(10, 10), linestyle="solid", target=None, ncol=2,
                     bbox_to_anchor=(0.0, 1.0)):
        """Plot a scatter corner plot of all accepted parameter pair values

        Plots scatter plots for parameter values in 2D subplots (and optionally
        the histogram of points in the 1D diagonal subplots).

        Parameters
        ----------
        ax : list of matplotlib axes or None, default=None
            If ax is None a new figure and axes are created to make a corner
            plot otherwise a set of axes are formatted correctly for a corner
            plot. If existing ax is not correctly shaped there will be an error
        ranges : list, default=None
            The list of ranges to set the number of rows and columns (if this
            is None the ranges set on initialisation will be used)
        points : float(n_targets, n_points, {n_params} or {n_summaries})
            The points to scatter plot, if None the class instance of
            `parameters.accepted` will be used
        label : str or None, default=None
            Name to be used in the legend
        axis_labels : list of str or None, default=None
            A list of names for each parameter, no axis labels if None
        colours : str or list or None, default=None
            The colour or list of colours to use for the different targets, if
            None then the normal matplotlib colours are used
        hist : bool, default=True
            Whether or not to plot histograms on the diagonal of the corner
            plot
        s : float, default=5.
            The size of the marker points in the scatter plot
        alpha : float, default=1.
            The amount of alpha colour for the marker points
        figsize : tuple, default=(10, 10)
            The size (in inches) to create a figure (if ax is None)
        linestyle : str, default="solid"
            Linestyle for the histograms
        target : None or int or list, default=None
            The indices to choose which target's points are plotted
        format : bool, default=False
            If formatting is not needed
        ncols : int, default=2
            Number of columns for the legend
        bbox_to_anchor : tuple, default=(0.0, 1.0)
            Position to fix the legend to

        Returns
        -------
        axes object:
            The scatter plot axes
        """
        if ranges is None:
            ranges = self.ranges
        if points is None:
            points = self.parameters.accepted
        return self._scatter_plot(
            ax=ax, ranges=ranges, points=points, label=label,
            axis_labels=axis_labels, colours=colours, hist=hist, s=s,
            alpha=alpha, figsize=figsize, linestyle=linestyle, target=target,
            ncol=ncol, bbox_to_anchor=bbox_to_anchor)

    def scatter_summaries(self, ax=None, ranges=None, points=None, label=None,
                          axis_labels=None, colours=None, hist=True, s=5,
                          alpha=1., figsize=(10, 10), linestyle="solid",
                          gridsize=100, target=None, format=False, ncol=2,
                          bbox_to_anchor=(0.0, 1.0)):
        """Plot a scatter corner plot of all accepted summary pair values

        Plots scatter plots for summary values in 2D subplots (and optionally
        the histogram of points in the 1D diagonal subplots).

        Parameters
        ----------
        ax : list of matplotlib axes or None, default=None
            If ax is None a new figure and axes are created to make a corner
            plot otherwise a set of axes are formatted correctly for a corner
            plot. If existing ax is not correctly shaped there will be an error
        ranges : list, default=None
            The list of ranges to set the number of rows and columns (if this
            is None the ranges set on initialisation will be used)
        points : float(n_targets, n_points, {n_params} or {n_summaries})
            The points to scatter plot, if None the class instance of
            `summaries.accepted` will be used
        label : str or None, default=None
            Name to be used in the legend
        axis_labels : list of str or None, default=None
            A list of names for each parameter, no axis labels if None
        colours : str or list or None, default=None
            The colour or list of colours to use for the different targets, if
            None then the normal matplotlib colours are used
        hist : bool, default=True
            Whether or not to plot histograms on the diagonal of the corner
            plot
        s : float, default=5.
            The size of the marker points in the scatter plot
        alpha : float, default=1.
            The amount of alpha colour for the marker points
        figsize : tuple, default=(10, 10)
            The size (in inches) to create a figure (if ax is None)
        linestyle : str, default="solid"
            Linestyle for the histograms
        target : None or int or list, default=None
            The indices to choose which target's points are plotted
        format : bool, default=False
            If formatting is not needed
        ncols : int, default=2
            Number of columns for the legend
        bbox_to_anchor : tuple, default=(0.0, 1.0)
            Position to fix the legend to

        Returns
        -------
        axes object:
            The scatter plot axes

        Todo
        ----
        There are many extra matplotlib parameters which could be passed,
        although this is not done because the axis is returned which can then
        be manipulated.
        """
        if points is None:
            points = self.summaries.accepted
            n_summaries = self.n_summaries
        else:
            n_summaries = points[0].shape[-1]
        gridsize = self.get_gridsize(gridsize, n_summaries)
        targets, n_targets = self.target_choice(target)
        if ranges is None:
            ranges = [
                np.linspace(
                    np.min(
                        np.array([
                            np.min(points[target][:, summary])
                            for target in targets])),
                    np.max(
                        np.array([
                            np.max(points[target][:, summary])
                            for target in targets])),
                    gridsize[summary])
                for summary in range(n_summaries)]
        return self._scatter_plot(
            ax=ax, ranges=ranges, points=points, label=label,
            axis_labels=axis_labels, colours=colours, hist=hist, s=s,
            alpha=alpha, figsize=figsize, linestyle=linestyle, target=target,
            format=format, ncol=ncol, bbox_to_anchor=bbox_to_anchor)

    def F_distance(self, x, y, F):
        """Default distance measure, squared difference scaled by Fisher matrix

        The target summaries are expanded on the zeroth axis so that it can
        broadcast with the simulation summaries.

        Parameters
        ----------
        x : (any, n_params)
            Envisioned as a whole batch of summaries of simulations
        y : (n_params)
            Envisioned as a target summary
        F : (n_params)
            Fisher information matrix to scale the summary direction
        """
        difference = x - np.expand_dims(y, 0)
        return np.sqrt(
            np.einsum(
                "ik,kl,il->i",
                difference,
                F,
                difference))

    def euclidean_distance(self, x, y, aux=None):
        """Euclidean distance between simulation summaries and target summary

        The target summaries are expanded on the zeroth axis so that it can
        broadcast with the simulation summaries.

        Parameters
        ----------
        x : (any, n_params)
            Envisioned as a whole batch of summaries of simulations
        y : (n_params)
            Envisioned as a target summary
        aux : None, default=Note
            Empty holder so that function works in the same way as `F_distance`
        """
        difference = x - np.expand_dims(y, 0)
        return np.sqrt(np.sum(np.square(difference), -1))



import pytest
import jax
import jax.numpy as np
import tensorflow as tf
from test.aggregated_defaults import aggregatedTests
from test.numerical_gradient_defaults import numericalGradientTests
from imnn.imnn import DatasetNumericalGradientIMNN


class datasetNumericalGradientTests(
        aggregatedTests, numericalGradientTests):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.kwargs["fiducial"] = self.fiducial.reshape(
            (self.n_devices,
             self.n_s // (self.n_devices * self.n_per_device),
             self.n_per_device)
            + self.input_shape)

        self.kwargs["derivative"] = self.derivative.reshape(
            (self.n_devices,
             2 * self.n_params * self.n_d //
                (self.n_devices * self.n_per_device),
             self.n_per_device)
            + self.input_shape)

        self.kwargs["validation_fiducial"] = self.validation_fiducial.reshape(
            (self.n_devices,
             self.n_s // (self.n_devices * self.n_per_device),
             self.n_per_device)
            + self.input_shape)

        self.kwargs["validation_derivative"] = \
            self.validation_derivative.reshape(
                (self.n_devices,
                 2 * self.n_params * self.n_d //
                    (self.n_devices * self.n_per_device),
                 self.n_per_device)
                + self.input_shape)

        self.reduced_kwargs["fiducial"] = self.fiducial.reshape(
            (self.n_devices,
             self.n_s // (self.n_devices * self.n_per_device),
             self.n_per_device)
            + self.input_shape)

        self.reduced_kwargs["derivative"] = \
            self.reduced_derivative.reshape(
                (self.n_devices,
                 2 * self.n_params * self.reduced_n_d //
                    (self.n_devices * self.n_per_device),
                 self.n_per_device)
                + self.input_shape)

        self.reduced_kwargs["validation_fiducial"] = \
            self.validation_fiducial.reshape(
                (self.n_devices,
                 self.n_s // (self.n_devices * self.n_per_device),
                 self.n_per_device)
                + self.input_shape)

        self.reduced_kwargs["validation_derivative"] = \
            self.reduced_validation_derivative.reshape(
                (self.n_devices,
                 2 * self.n_params * self.reduced_n_d //
                    (self.n_devices * self.n_per_device),
                 self.n_per_device)
                + self.input_shape)

        self.arrays = ["θ_fid", "δθ"]

    def preload(self, dictionary, state=False, validate=False):
        dictionary["host"] = jax.devices("cpu")[0]
        dictionary["devices"] = jax.devices()
        dictionary["fiducial"] = [
            tf.data.Dataset.from_tensor_slices(
                fid).repeat().as_numpy_iterator()
            for fid in dictionary["fiducial"]]
        dictionary["derivative"] = [
            tf.data.Dataset.from_tensor_slices(
                der).repeat().as_numpy_iterator()
            for der in dictionary["derivative"]]
        if state:
            dictionary["key_or_state"] = self.state
        if (not self.simulate) and (not validate):
            dictionary.pop("validation_fiducial")
            dictionary.pop("validation_derivative")
        else:
            dictionary["validation_fiducial"] = [
                tf.data.Dataset.from_tensor_slices(
                    fid).repeat().as_numpy_iterator()
                for fid in dictionary["validation_fiducial"]]
            dictionary["validation_derivative"] = [
                tf.data.Dataset.from_tensor_slices(
                    der).repeat().as_numpy_iterator()
                for der in dictionary["validation_derivative"]]

        return dictionary

    def splitting_resize(self, _kwargs):
        _kwargs["fiducial"] = None
        _kwargs["derivative"] = None
        if "validation_fiducial" in _kwargs.keys():
            _kwargs["validation_fiducial"] = None
            _kwargs["validation_derivative"] = None
        return _kwargs


test = datasetNumericalGradientTests(
    imnn=DatasetNumericalGradientIMNN,
    filename="dataset_numerical_gradient",
    n_per_device=100)


@pytest.mark.parametrize("kwargs", [test.kwargs, test.reduced_kwargs])
@pytest.mark.parametrize("state", [True, False])
@pytest.mark.parametrize("validate", [True, False])
@pytest.mark.parametrize(
    "input_variable",
    [None, list(), 1., 1, np.zeros((1,)), test.rng, tuple(), (0, 0),
     (test.model[0], 0), test.bad_model, test.state])
@pytest.mark.parametrize("variable", test.kwargs.keys())
def test_initialisation_parameters_(
        variable, kwargs, input_variable, state, validate):
    test.initialise_parameters(
        variable, kwargs, input_variable, state=state, validate=validate)


@pytest.mark.parametrize("validate", [True, False])
@pytest.mark.parametrize("state", [False, True])
@pytest.mark.parametrize("variable", ["n_s", "n_d", "same"])
def test_splitting_(variable, validate, state):
    test.splitting(variable, test.kwargs, state=state, validate=validate)


@pytest.mark.parametrize("kwargs", [test.kwargs, test.reduced_kwargs])
@pytest.mark.parametrize("state", [True, False])
@pytest.mark.parametrize("validate", [True, False])
@pytest.mark.parametrize(
    "input_variable", [None, list(), 1., 1, np.zeros((1,)), test.rng])
@pytest.mark.parametrize("variable", test.fit_kwargs.keys())
def test_fit_parameters_(variable, kwargs, input_variable, state, validate):
    test.fit_parameters(
        variable, kwargs, test.fit_kwargs, input_variable, state=state,
        validate=validate)


@pytest.mark.parametrize("state", [True, False])
@pytest.mark.parametrize("validate", [True, False])
@pytest.mark.parametrize("fit", [True, False])
@pytest.mark.parametrize("none_first", [True, False])
@pytest.mark.parametrize("kwargs", [test.kwargs, test.reduced_kwargs])
def test_combined_running_test_(kwargs, state, validate, fit, none_first):
    test.combined_running_test(
        [test.single_target_data, test.batch_target_data], kwargs,
        test.fit_kwargs, state=state, validate=validate, fit=fit,
        none_first=none_first, aggregated=True)



import inspect
import jax
import jaxlib
import jax.numpy as np
import numpy as onp
from jax.tree_util import tree_flatten, tree_unflatten


def _check_boolean(input, name):
    """Exception raising if input is not a boolean

    Checks whether input is not ``None`` and if not checks that the input is a
    boolean.

    Parameters
    ----------
    input: any
        The input parameter to be checked
    name: str
        The name of the variable for printing explicit errors in ``Exception``

    Returns
    -------
    bool
        Returns the input if all checks pass

    Raises
    ------
    ValueError
        If input is None
    TypeError
        If input does not have correct type
    """
    if input is None:
        raise ValueError(f"`{name}` is None")
    if not ((input is True) or (input is False)):
        raise TypeError(f"`{name}` must be type {bool} but is " +
                        f"{type(input)}")
    return input


def _check_type(input, target_type, name, shape=None, allow_None=False):
    """Exception raising checks for parameter types (and shapes)

    Checks whether input is not ``None`` and if not checks that the input has
    the specified type and if not raises a warning.

    Also can check the shape of objects, although for jax numpy arrays this is
    done with :func:~`imnn.utils.check_input` since arrays can have various
    types and it is hard to predict which depending on case.

    Parameters
    ----------
    input: any
        The input parameter to be checked
    target_type: type
        The target type that the input parameter should be an instance of
    name: str
        The name of the variable for printing explicit errors in ``Exception``
    shape: tuple or int or None, default=None
        The length of a list or tuple or the shape of an array if also checked
    allow_None: bool, default=False
        Whether a ``None`` input can be returned as None without raising error

    Returns
    -------
    any
        Returns the input if all checks pass

    Raises
    ------
    ValueError
        If input is None
    ValueError
        If input shape is incorrect
    TypeError
        If input does not have correct type
    """
    if (input is None) and (not allow_None):
        raise ValueError(f"`{name}` is None")
    elif (input is None) and allow_None:
        return input
    elif not isinstance(input, target_type):
        raise TypeError(f"`{name}` must be type {target_type} but is " +
                        f"{type(input)}")
    elif shape is not None:
        if hasattr(input, "shape"):
            input_shape = input.shape
            string = "shape"
        else:
            input_shape = len(input)
            string = "length"
        if input_shape != shape:
            raise ValueError(f"`{name}` must have shape {shape} but has " +
                             f"{string} {input_shape}.")
    return input


def _check_input(input, shape, name, allow_None=False):
    """Exception raising checks for numpy array shapes

    Checks whether input is not ``None`` and if not checks that the input is a
    jax numpy array and if not raises a warning. If the input is a jax numpy
    array it then checks the shape is the same as the required shape.

    Can also allow ``None`` to be passed if it input is not essential.

    Parameters
    ----------
    input: any
        The input parameter to be checked
    shape: tuple
        The shape that the input is required to be
    name: str
        The name of the variable for printing explicit errors in ``Exception``
    allow_None: bool, default=False
        Whether a ``None`` input can be returned as None without raising error

    Returns
    -------
    array
        Returns the input if all checks pass

    Raises
    ------
    ValueError
        If input is None
    ValueError
        If input shape is incorrect
    TypeError
        If input is not a jax array
    """
    if (input is None) and (not allow_None):
        raise ValueError(f"`{name}` is None")
    elif (input is None) and allow_None:
        return input
    elif not isinstance(
            input,
            (jax.interpreters.xla.device_array, np.ndarray, onp.ndarray)):
        raise TypeError(f"`{name}` must be a jax array")
    else:
        if input.shape != shape:
            raise ValueError(f"`{name}` should have shape {shape} but has " +
                             f"{input.shape}")
    return input


def _check_devices(devices):
    """Exception raising checks for XLA devices

    Checks whether input is not ``None`` and if not checks that the input is a
    list of XLA devices

    Parameters
    ----------
    devices: list
        The list of XLA devices (from ``jax.devices()`` for example)

    Returns
    -------
    list
        The list of XLA devices (from ``jax.devices()`` for example)

    Raises
    ------
    ValueError
        If input is None
    ValueError
        If input shape is incorrect
    TypeError
        If inputs are not XLA devices
    """
    if (devices is None):
        raise ValueError("`devices` is None")
    elif not isinstance(devices, list):
        raise TypeError(f"`devices` must be type {list} but is " +
                        f"{type(devices)}")
    elif len(devices) < 1:
        raise ValueError("`devices` has no elements in")
    elif not all([
            isinstance(device, jaxlib.xla_extension.Device)
            for device in devices]):
        raise TypeError("all elements of `devices` must be xla devices")
    return devices


def _check_host(host):
    """Exception raising checks for XLA devices

    Checks whether input is not ``None`` and if not checks that the input is an
    XLA devices

    Parameters
    ----------
    devices: xla device
        the XLA device (from ``jax.devices()[0]`` for example)

    Returns
    -------
    xla device
        the XLA device (from ``jax.devices()[0]`` for example)

    Raises
    ------
    ValueError
        If input is None
    TypeError
        If inputs are not XLA devices
    """
    if (host is None):
        raise ValueError("`host` is None")
    elif not isinstance(host, jaxlib.xla_extension.Device):
        raise TypeError(f"`host` must be an xla device but is a {type(host)}")
    return host


def _check_model(model):
    """Exception raising for ``jax.experimental.stax``-like model

    Checks model is not ``None`` and if not checks that the input is  a tuple
    with two functions which take the correct number of inputs. For more
    information see `jax.experimental.optimizers <https://jax.readthedocs.io/e
    n/stable/jax.experimental.stax.html>`_.

    Parameters
    ----------
    model: any
        The model to be checked

    Returns
    -------
    tuple
        Returns the input if all checks pass

    Raises
    ------
    ValueError
        If input is None
    ValueError
        If optimiser is not a tuple of functions
    ValueError
        If element functions do not require the correct number of inputs
    TypeError
        If input is not a tuple of functions
    TypeError
        If elements of tuples are not functions
    """
    if model is None:
        raise ValueError("`model` is None")
    elif not isinstance(model, tuple):
        raise TypeError(f"`model` must be type {tuple} but is {type(model)}")
    else:
        if len(model) != 2:
            raise ValueError("`model` must be a tuple of two functions. The " +
                             "first for initialising the model and the " +
                             "second to call the model")
        try:
            inspect.signature(model[0]).parameters
        except Exception:
            raise TypeError("first element of `model` must take two arguments")
        if len(inspect.signature(model[0]).parameters) != 2:
            raise ValueError(
                "first element of `model` must take two arguments")
        try:
            inspect.signature(model[1]).parameters
        except Exception:
            raise TypeError(
                "second element of `model` must take three arguments")
        if len(inspect.signature(model[1]).parameters) != 3:
            raise ValueError(
                "second element of `model` must take three arguments")
    return model


def _check_model_output(output_shape, expected_shape):
    """Exception raising for shape of model output

    Parameters
    ----------
    output_shape: tuple
        Shape of the model output
    expected_shape: tuple
        Desired shape for the model output

    Raises
    ------
    ValueError
        If input does not equal the expected shape
    """
    if output_shape != expected_shape:
        raise ValueError("`model` outputs should have shape " +
                         f"{expected_shape} but is {output_shape}")


def _check_optimiser(optimiser):
    """Exception raising for ``jax.experimental.optimizers``-like optimiser

    Checks optimiser is not ``None`` and if not checks that the input is a
    ``jax.experimental.optimizers``-like instance. To allow more freedom on
    personal optimisers, if the optimiser is not an instance of a
    ``jax.experimental.optimizers`` object then it checks that the optimiser is
    at least a tuple with three functions which take the correct number of
    inputs. For more information see `jax.experimental.optimizers <https://jax.
    readthedocs.io/en/stable/jax.experimental.optimizers.html>`_.

    Parameters
    ----------
    optimiser: any
        The optimiser to be checked

    Returns
    -------
    tuple or jax.experimental.optimizers-like object
        ``jax.experimental.optimizers``-like optimiser for initialising state,
        updating parameters and state and getting parameters from state

    Raises
    ------
    ValueError
        If input is None
    ValueError
        If optimiser is not a tuple of functions or an instance of a
        ``jax.experimental.optimizers`` object
    ValueError
        If element functions do not require the correct number of inputs
    TypeError
        If input is not a tuple of functions
    TypeError
        If elements of tuples are not functions
    """
    string = "`optimiser` must be tuple of three functions. The first for " + \
        "initialising the state, the second to update the state and the " + \
        "third to get parameters from the state."
    if optimiser is None:
        raise ValueError("`optimiser` is None")
    if isinstance(optimiser, jax.experimental.optimizers.Optimizer):
        return optimiser
    else:
        try:
            length = len(optimiser)
        except Exception:
            raise TypeError(string)
        if length != 3:
            raise TypeError(string)
        try:
            inspect.signature(optimiser[0]).parameters
        except Exception:
            raise TypeError(
                "first element of `optimiser` must take one argument")
        if len(inspect.signature(optimiser[0]).parameters) != 1:
            raise ValueError(
                "first element of `optimiser` must take one argument")
        try:
            inspect.signature(optimiser[1]).parameters
        except Exception:
            raise TypeError(
                "second element of `optimiser` must take three arguments")
        if len(inspect.signature(optimiser[1]).parameters) != 3:
            raise ValueError(
                "second element of `optimiser` must take three arguments")
        try:
            inspect.signature(optimiser[2]).parameters
        except Exception:
            raise TypeError(
                "third element of `optimiser` must take one argument")
        if len(inspect.signature(optimiser[2]).parameters) != 3:
            raise ValueError(
                "third element of `optimiser` must take one argument")
        return optimiser


def _check_state(state):
    """RNG and state checking for ``jax.experimental.optimizers``-like state

    Checks state is not ``None`` and if not checks that the input is a random
    number generator or not. If the shape and type is correct for a random
    number generator then that is passed back to initialise a new state and
    network parameters. If the input is not a random number generator then the
    state is returned. Note that no specific checks are made for this state
    but expect an error in ``imnn.imnn._imnn._IMNN().fit()`` if the parameters
    cannot be obtained from the state.

    Parameters
    ----------
    state: any
        The state to be checked

    Returns
    -------
    None, int(2,) or ``jax.experimental.optimizers``-like state, None
        Returns either a random number generator or a state depending on input

    Raises
    ------
    ValueError
        If input is None
    """
    string = "`state` not a jax optimiser state - attempting to use it anyway"
    if (state is None):
        raise ValueError("`key_or_state` is None")
    elif isinstance(
            state,
            (jax.interpreters.xla.device_array, np.ndarray, onp.ndarray)):
        if state.shape == (2,):
            return None, state
        else:
            print(string)
            return state, None
    elif isinstance(state, jax.experimental.optimizers.OptimizerState):
        return state, None
    else:
        print(string)
        return state, None


def _check_simulator(simulator):
    """Checks that the simulator is callable and takes just two parameters

    For the simulator to be functional within the IMNN it must have a specific
    form where it takes a random number and a set of parameter values and
    returns a simulation generated at those parameter values.

    Parameters
    ----------
    simulator: any
        the simulator to be checked

    Returns
    -------
    fn:
        the simulator function

    Raises
    ------
    ValueError
        if simlator is None or it is a function which takes the wrong number of
        parameters
    TypeError
        if simulator is not a callable function
    """
    if simulator is None:
        raise ValueError("`simulator` is None")
    elif not callable(simulator):
        raise TypeError("`simulator` must be type <class 'function'> but is " +
                        f"{type(simulator)}")
    else:
        if len(inspect.signature(simulator).parameters) != 2:
            raise ValueError("`simulator` must take two arguments, a " +
                             "JAX prng and simulator parameters.")
        return simulator


def _check_splitting(size, name, n_devices, n_per_device):
    """Checks whether integers are exactly divisible by two other integers

    To be able to run the aggregated IMNNs, datasets have to be reshaped across
    different devices. Since ``n_s`` and ``n_d`` are known and fixed, it is
    possible to check that the reshaping can be done and return an error if the
    splitting is not possible.

    Parameters
    ----------
    size: int
        The number (of simulations) to be split over devices
    name: str
        The variable quantity name for flagging the error
    n_devices: int
        One of the numbers to be divided by (ostensibly the number of jax
        devices available to be aggregated over)
    n_per_device: int
        The other number to be divided by (the number of simulations to run in
        a batch on each device)

    Raises
    ------
    ValueError
        if the number of simulations is not exactly divisible by the number of
        devices and the number within a batch
    """
    if (size / (n_devices * n_per_device)
            != float(size // (n_devices * n_per_device))):
        raise ValueError(f"`{name}` of {size} will not split evenly between " +
                         f"{n_devices} devices when calculating " +
                         f"{n_per_device} per device.")


def _check_statistics_set(invF, dμ_dθ, invC, μ):
    """Checks that quantities needed for IMNN score compression are assigned

    Parameters
    ----------
    invF: float(n_params, n_params) or None
        The inverse of the Fisher information matrix
    dμ_dθ: float(n_summaries, n_params) or None
        The derivative of the mean of the network outputs with respect to model
        parameters
    invC: float(n_summaries, n_summaries) or None
        The inverse of the covariance of the network outputs
    μ: float(n_summaries) or None
        The mean of the network outputs

    Raises
    ------
    ValueError
        if any of the inputs are None
    """
    if (invF is None) or (dμ_dθ is None) or (invC is None) or (μ is None):
        raise ValueError(
            "Fisher information has not yet been calculated. Please run " +
            "`imnn.set_F_statistics({w}, {key}, {validate}) with " +
            "`w = imnn.final_w`, `w = imnn.best_w`, `w = imnn.inital_w` " +
            "or otherwise, `validate = True` should be set if not " +
            "simulating on the fly.")


def add_nested_pytrees(*pytrees):
    """Sums together any number of identically shaped pytrees

    All pytrees are flattened then summed and then reshaped back into their
    tree

    Parameters
    ----------
    pytrees: any number of pytrees
        the pytrees to be summed together

    Returns
    -------
    pytree:
        the single pytree resulting from summing together all input pytrees

    Todo
    ----
    No checking for identical shape is done, although it should be done really
    """
    value_flat, value_tree = tree_flatten(pytrees[0])
    for pytree in pytrees[1:]:
        next_value_flat, _ = tree_flatten(pytree)
        value_flat = list(map(sum, zip(value_flat, next_value_flat)))
    return tree_unflatten(value_tree, value_flat)



from imnn.imnn import AggregatedGradientIMNN
from imnn.utils.utils import _check_type


class DatasetGradientIMNN(AggregatedGradientIMNN):
    """Information maximising neural network fit using known derivatives


    The outline of the fitting procedure is that a set of :math:`i\\in[1, n_s]`
    simulations :math:`{\\bf d}^i` originally generated at fiducial model
    parameter :math:`{\\bf\\theta}^\\rm{fid}`, and their derivatives
    :math:`\\partial{\\bf d}^i/\\partial\\theta_\\alpha` with respect to
    model parameters are used. The fiducial simulations, :math:`{\\bf d}^i`,
    are passed through a network to obtain summaries, :math:`{\\bf x}^i`, and
    the jax automatic derivative of these summaries with respect to the inputs
    are calculated :math:`\\partial{\\bf x}^i\\partial{\\bf d}^j\\delta_{ij}`.
    The chain rule is then used to calculate

    .. math::
        \\frac{\\partial{\\bf x}^i}{\\partial\\theta_\\alpha} =
        \\frac{\\partial{\\bf x}^i}{\\partial{\\bf d}^j}
        \\frac{\\partial{\\bf d}^j}{\\partial\\theta_\\alpha}

    With :math:`{\\bf x}^i` and
    :math:`\\partial{{\\bf x}^i}/\\partial\\theta_\\alpha` the covariance

    .. math::
        C_{ab} = \\frac{1}{n_s-1}\\sum_{i=1}^{n_s}(x^i_a-\\mu^i_a)
        (x^i_b-\\mu^i_b)

    and the derivative of the mean of the network outputs with respect to the
    model parameters

    .. math::
        \\frac{\\partial\\mu_a}{\\partial\\theta_\\alpha} = \\frac{1}{n_d}
        \\sum_{i=1}^{n_d}\\frac{\\partial{x^i_a}}{\\partial\\theta_\\alpha}

    can be calculated and used form the Fisher information matrix

    .. math::
        F_{\\alpha\\beta} = \\frac{\\partial\\mu_a}{\\partial\\theta_\\alpha}
        C^{-1}_{ab}\\frac{\\partial\\mu_b}{\\partial\\theta_\\beta}.

    The loss function is then defined as

    .. math::
        \\Lambda = -\\log|{\\bf F}| + r(\\Lambda_2) \\Lambda_2

    Since any linear rescaling of a sufficient statistic is also a sufficient
    statistic the negative logarithm of the determinant of the Fisher
    information matrix needs to be regularised to fix the scale of the network
    outputs. We choose to fix this scale by constraining the covariance of
    network outputs as

    .. math::
        \\Lambda_2 = ||{\\bf C}-{\\bf I}|| + ||{\\bf C}^{-1}-{\\bf I}||

    Choosing this constraint is that it forces the covariance to be
    approximately parameter independent which justifies choosing the covariance
    independent Gaussian Fisher information as above. To avoid having a dual
    optimisation objective, we use a smooth and dynamic regularisation strength
    which turns off the regularisation to focus on maximising the Fisher
    information when the covariance has set the scale

    .. math::
        r(\\Lambda_2) = \\frac{\\lambda\\Lambda_2}{\\Lambda_2-\\exp
        (-\\alpha\\Lambda_2)}.

    To enable the use of large data (or networks) the whole procedure is
    aggregated. This means that the passing of the simulations through the
    network is farmed out to the desired XLA devices, and recollected,
    ``n_per_device`` inputs at a time. These are then used to calculate the
    automatic gradient of the loss function with respect to the calculated
    summaries and derivatives, :math:`\\partial\\Lambda/\\partial{\\bf x}^i`
    (which is a fairly small computation as long as ``n_summaries`` and ``n_s``
    {and ``n_d``} are not huge). Once this is calculated, the simulations are
    passed through the network AGAIN this time calculating the Jacobian of the
    network output with respect to the network parameters
    :math:`\\partial{\\bf x}^i/\\partial{\\bf w}` which is then combined via
    the chain rule to get

    .. math::
        \\frac{\\partial\\Lambda}{\\partial{\\bf w}} =
        \\frac{\\partial\\Lambda}{\\partial{\\bf x}^i}
        \\frac{\\partial{\\bf x}^i}{\\partial{\\bf w}}

    This can then be passed to the optimiser.

    In DatasetGradientIMNN the input datasets should be lists of ``n_devices``
    ``tf.data.Datasets``. Please note, due to the many various ways of
    constructing datasets to load data, there is no checking and any improperly
    made dataset will either fail (best result) or provide the wrong result
    (worst case scenario!). For this reason it is advised to use
    :func:`~AggregatedGradientIMNN` if data will fit into CPU memory at least.
    If not, the next safest way is to construct a set of TFRecords and
    construct the dataset from that.

    Examples
    --------
    Here are various ways to construct the datasets for passing to the
    :func:~`DatasetGradientIMNN`. Note that these are not the only
    ways, but they should give something to follow to generate your own
    datasets. First we'll generate some data (just random noise with zero mean
    and unit variance). We'll generate 1000 simulations at the fiducial and
    we'll use jax to calculate the derivatives with respect to the mean and
    variance for 100 of these. We'll save each of these simulations into its
    own individual file (named by seed value).

    .. code-block:: python

        import glob
        import jax
        import jax.numpy as np
        import tensorflow as tf
        from imnn import TFRecords
        from functools import partial
        from imnn.utils import value_and_jacfwd

        n_s = 1000
        n_d = 100
        n_params = 2
        input_shape = (10,)

        def simulator(key, θ):
            return θ[0] + (jax.random.normal(key, shape=input_shape)
                * np.sqrt(θ[1]))

        θ_fid = np.array([0., 1.])

        get_sims_and_ders = value_and_jacfwd(simulator, argnums=1)

        rng = jax.random.PRNGKey(0)
        rng, data_key = jax.random.split(rng)
        data_keys = np.array(jax.random.split(rng, num=2 * n_s))

        fiducial, derivative = jax.vmap(get_sims_and_ders)(
            data_keys[:n_d], np.repeat(np.expand_dims(θ_fid, 0), n_d, axis=0))

        remaining = jax.vmap(simulator)(
            data_keys[n_d:n_s],
            np.repeat(np.expand_dims(θ_fid, 0), n_s - n_d, axis=0))

        validation_fiducial, validation_derivative = jax.vmap(
            get_sims_and_ders)(
                data_keys[n_s:n_s + n_d],
                np.repeat(np.expand_dims(θ_fid, 0), n_d, axis=0))

        validation_remaining = jax.vmap(simulator)(
            data_keys[n_s + n_d:],
            np.repeat(np.expand_dims(θ_fid, 0), n_s - n_d, axis=0))

        for i, (simulation, validation_simulation) in enumerate(zip(
                fiducial, validation_fiducial)):
            np.save(f"tmp/fiducial_{i:04d}.npy", simulation)
            np.save(f"tmp/validation_fiducial_{i:04d}.npy",
                    validation_simulation)

        for i, (simulation, validation_simulation) in enumerate(zip(
                derivative, validation_derivative)):
            np.save(f"tmp/derivative_{i:04d}.npy", simulation)
            np.save(f"tmp/validation_derivative_{i:04d}.npy",
                    validation_simulation)

    Now we'll define how many devices to farm out our calculations to. We need
    to know this because we want to make a separate dataset for each device.
    We'll also set the number of simulations which can be processed at once on
    each device, this should be as high as possible without running out of
    memory on any individual device for quickest fitting

    .. code-block:: python

        devices = jax.devices("gpu")
        n_devices = len(devices)
        n_per_device = 100

    To best accelerate the aggregation of the gradient calculation the
    computation is split into two parts, a ``main`` loop which loops through
    ``n_d`` simulation with its derivative with respect to model parameters,
    and a ``remaining`` loop of ``n_s - n_d`` iterations, where just
    simulations are looped through to calculate any other necessary summaries
    to estimate the covariance. Note this is true even if ``n_s = n_s`` the
    remaining loop just has zero iterations. So to construct the dataset define
    the shapes for the data to be reshaped into for proper construction of the
    datasets to be used when fitting the IMNN.

    .. code-block:: python

        batch_shape = (
            n_devices,
            n_d // (n_devices * n_per_device),
            n_per_device) + input_shape

        remaining_batch_shape = (
            n_devices,
            (n_s - n_d) // (n_devices * n_per_device),
            n_per_device) + input_shape

    The simplest way to construct a dataset is simply using the numpy arrays
    from memory (note if you're going to do this you should really just use
    ``AggregatedGradientIMNN``, its more or less the same!), i.e.

    .. code-block:: python

        main = [
            tf.data.Dataset.from_tensor_slices(
                (fiducial, derivative)).repeat().as_numpy_iterator()
            for fiducial, derivative in zip(
                fiducial.reshape(batch_shape),
                derivative.reshape(batch_shape + (n_params,)))]

        remaining = [
            tf.data.Dataset.from_tensor_slices(fiducial
                ).repeat().as_numpy_iterator()
            for fiducial in remaining.reshape(
                remaining_batch_shape)]

        validation_main = [
            tf.data.Dataset.from_tensor_slices(
                (fiducial, derivative)).repeat().as_numpy_iterator()
            for fiducial, derivative in zip(
                validation_fiducial.reshape(batch_shape),
                validation_derivative.reshape(batch_shape + (n_params,)))]

        validation_remaining = [
            tf.data.Dataset.from_tensor_slices(fiducial
                ).repeat().as_numpy_iterator()
            for fiducial in validation_remaining.reshape(
                remaining_batch_shape)]

    However, if the data is too large to fit in memory then we can use the
    npy files that we saved by loading them via a generator

    .. code-block:: python

        def generator(directory, filename, total):
            i = 0
            while i < total:
                yield np.load(f"{directory}/{filename}_{i:04d}.npy")
                i += 1

    We can then build the datasets like:

    .. code-block:: python

        main = [
            tf.data.Dataset.zip((
                 tf.data.Dataset.from_generator(
                     partial(
                         generator,
                         "tmp",
                         "fiducial",
                         n_d),
                     tf.float32),
                tf.data.Dataset.from_generator(
                     partial(
                         generator,
                         "tmp",
                         "derivative",
                         n_d),
                     tf.float32))
                ).take(n_d // n_devices
                ).batch(n_per_device
                ).repeat(
                ).as_numpy_iterator()
            for _ in range(n_devices)]

        remaining = [
            tf.data.Dataset.from_generator(
                partial(
                    generator,
                    "tmp",
                    "remaining",
                    n_s - n_d),
                tf.float32
                ).take((n_s - n_d) // n_devices
                ).batch(n_per_device
                ).repeat(
                ).as_numpy_iterator()
            for _ in range(n_devices)]

        validation_main = [
            tf.data.Dataset.zip((
                 tf.data.Dataset.from_generator(
                     partial(
                         generator,
                         "tmp",
                         "validation_fiducial",
                         n_d),
                     tf.float32),
                tf.data.Dataset.from_generator(
                     partial(
                         generator,
                         "tmp",
                         "validation_derivative",
                         n_d),
                     tf.float32))
                ).take(n_d // n_devices
                ).batch(n_per_device
                ).repeat(
                ).as_numpy_iterator()
            for _ in range(n_devices)]

        validation_remaining = [
            tf.data.Dataset.from_generator(
                partial(
                    generator,
                    "tmp",
                    "validation_remaining",
                    n_s - n_d),
                tf.float32
                ).take((n_s - n_d) // n_devices
                ).batch(n_per_device
                ).repeat(
                ).as_numpy_iterator()
            for _ in range(n_devices)]

    The datasets must be built exactly like this, with the taking and batching
    and repeating. The zipping of the main datasets is equal as important to
    pass both the fiducial simulation and its derivative at once, which is
    needed to calculate the final gradient. To prefetch and cache the
    loaded files we can add the extra steps in the datasets, e.g.

    .. code-block:: python

        main = [
            tf.data.Dataset.zip((
                 tf.data.Dataset.from_generator(
                     partial(
                         generator,
                         "tmp",
                         "fiducial",
                         n_d),
                     tf.float32),
                tf.data.Dataset.from_generator(
                     partial(
                         generator,
                         "tmp",
                         "derivative",
                         n_d),
                     tf.float32))
                ).take(n_d // n_devices
                ).batch(n_per_device
                ).cache(
                ).prefetch(tf.data.AUTOTUNE
                ).repeat(
                ).as_numpy_iterator()

    etc.

    This loading will be quite slow because the files need to be opened each
    time, but we can build TFRecords which are quicker to load. There is a
    writer able to do the correct format. The TFRecords should be a couple
    hundred Mb for best flow-through, so we can keep filling the record until
    this size is reached.

    .. code-block:: python

        record_size = 200 #Mb
        writer = TFRecords.TFRecords(record_size=record_size)

    We need a function which grabs single simulations from an array (or file)
    to add to the record

    .. code-block:: python

        def get_simulation(seed, directory=None, filename=None):
            return np.load(f"{directory}/{filename}_{seed:04d}.npy")

        writer.write_record(
            n_sims=n_d,
            get_simulation=lambda seed: get_simulation(
                seed, directory="tmp", filename="fiducial"),
            directory="tmp",
            filename="fiducial")

        writer.write_record(
            n_sims=n_s - n_d,
            get_simulation=lambda seed: get_simulation(
                seed, directory="tmp", filename="remaining"),
            directory="tmp",
            filename="remaining")

        writer.write_record(
            n_sims=n_d,
            get_simulation=lambda seed: get_simulation(
                seed, directory="tmp", filename="derivative"),
            directory="tmp",
            filename="derivative")

        writer.write_record(
            n_sims=n_d,
            get_simulation=lambda seed: get_simulation(
                seed, directory="tmp", filename="validation_fiducial"),
            directory="tmp",
            filename="validation_fiducial")

        writer.write_record(
            n_sims=n_s - n_d,
            get_simulation=lambda seed: get_simulation(
                seed, directory="tmp", filename="validation_remaining"),
            directory="tmp",
            filename="validation_remaining")

        writer.write_record(
            n_sims=n_d,
            get_simulation=lambda seed: get_simulation(
                seed, directory="tmp", filename="validation_derivative"),
            directory="tmp",
            filename="validation_derivative")

    We can then read these to a dataset using the parser from the TFRecords
    class mapping the format of the data to a 32-bit float

    .. code-block:: python

        fiducial = [
            tf.data.TFRecordDataset(
                    sorted(glob.glob("tmp/fiducial_*.tfrecords")),
                    num_parallel_reads=1
                ).map(writer.parser
                ).skip(i * n_s // n_devices
                ).take(n_s // n_devices)
            for i in range(n_devices)]

        main = [
            tf.data.Dataset.zip((
                fiducial[i],
                tf.data.TFRecordDataset(
                    sorted(glob.glob("tmp/derivative_*.tfrecords")),
                    num_parallel_reads=1).map(
                        lambda example: writer.derivative_parser(
                            example, n_params=n_params)))
                ).take(n_d // n_devices
                ).batch(n_per_device
                ).repeat(
                ).as_numpy_iterator()
            for i in range(n_devices)]

        remaining = [
            fiducial[i].skip(n_d // n_devices
                ).take((n_s - n_d) // n_devices
                ).batch(n_per_device
                ).repeat(
                ).as_numpy_iterator()
            for i in range(n_devices)]

        validation_fiducial = [
            tf.data.TFRecordDataset(
                    sorted(glob.glob("tmp/validation_fiducial_*.tfrecords")),
                    num_parallel_reads=1
                ).map(writer.parser
                ).skip(i * n_s // n_devices
                ).take(n_s // n_devices)
            for i in range(n_devices)]

        validation_main = [
            tf.data.Dataset.zip((
                validation_fiducial[i],
                tf.data.TFRecordDataset(
                    sorted(glob.glob("tmp/validation_derivative_*.tfrecords")),
                    num_parallel_reads=1).map(
                        lambda example: writer.derivative_parser(
                            example, n_params=n_params)))
                ).take(n_d // n_devices
                ).batch(n_per_device
                ).repeat(
                ).as_numpy_iterator()
            for i in range(n_devices)]

        validation_remaining = [
            validation_fiducial[i].skip(n_d // n_devices
                ).take((n_s - n_d) // n_devices
                ).batch(n_per_device
                ).repeat(
                ).as_numpy_iterator()
            for i in range(n_devices)]

    Parameters
    ----------
    main : list of tf.data.Dataset().as_numpy_iterators()
        The simulations generated at the fiducial model parameter values used
        for calculating the covariance of network outputs and their
        derivatives with respect to the physical model parameters
        (for fitting). These are served ``n_per_device`` at a time as a
        numpy iterator from a TensorFlow dataset.
    remaining : list of tf.data.Dataset().as_numpy_iterators()
        The ``n_s - n_d`` simulations generated at the fiducial model parameter
        values used for calculating the covariance of network outputs with a
        derivative counterpart (for fitting). These are served ``n_per_device``
        at a time as a numpy iterator from a TensorFlow dataset.
    validation_main : list of tf.data.Dataset().as_numpy_iterators()
        The simulations generated at the fiducial model parameter values used
        for calculating the covariance of network outputs and their
        derivatives with respect to the physical model parameters
        (for validation). These are served ``n_per_device`` at a time as a
        numpy iterator from a TensorFlow dataset.
    validation_remaining : list of tf.data.Dataset().as_numpy_iterators()
        The ``n_s - n_d`` simulations generated at the fiducial model parameter
        values used for calculating the covariance of network outputs with a
        derivative counterpart (for validation). Served ``n_per_device``
        at time as a numpy iterator from a TensorFlow dataset.
    n_remaining: int
        The number simulations where only the fiducial simulations are
        calculated. This is zero if ``n_s`` is equal to ``n_d``.
    n_iterations : int
        Number of iterations through the main summarising loop
    n_remaining_iterations : int
        Number of iterations through the remaining simulations used for quick
        loops with no derivatives
    batch_shape: tuple
        The shape which ``n_d`` should be reshaped to for aggregating.
        ``n_d // (n_devices * n_per_device), n_devices, n_per_device``
    remaining_batch_shape: tuple
        The shape which ``n_s - n_d`` should be reshaped to for aggregating.
        ``(n_s - n_d) // (n_devices * n_per_device), n_devices, n_per_device``
    """
    def __init__(self, n_s, n_d, n_params, n_summaries, input_shape, θ_fid,
                 model, optimiser, key_or_state, main, remaining, host,
                 devices, n_per_device, validation_main=None,
                 validation_remaining=None):
        """Constructor method

        Initialises all IMNN attributes, constructs neural network and its
        initial parameter values and creates history dictionary. Also fills the
        simulation attributes (and validation if available).

        Parameters
        ----------
        n_s : int
            Number of simulations used to calculate summary covariance
        n_d : int
            Number of simulations used to calculate mean of summary derivative
        n_params : int
            Number of model parameters
        n_summaries : int
            Number of summaries, i.e. outputs of the network
        input_shape : tuple
            The shape of a single input to the network
        θ_fid : float(n_params,)
            The value of the fiducial parameter values used to generate inputs
        model : tuple, len=2
            Tuple containing functions to initialise neural network
            ``fn(rng: int(2), input_shape: tuple) -> tuple, list`` and the
            neural network as a function of network parameters and inputs
            ``fn(w: list, d: float(None, input_shape)) -> float(None, n_summari
            es)``.
            (Essentibly stax-like, see `jax.experimental.stax <https://jax.read
            thedocs.io/en/stable/jax.experimental.stax.html>`_))
        optimiser : tuple, len=3
            Tuple containing functions to generate the optimiser state
            ``fn(x0: list) -> :obj:state``, to update the state from a list of
            gradients ``fn(i: int, g: list, state: :obj:state) -> :obj:state``
            and to extract network parameters from the state
            ``fn(state: :obj:state) -> list``.
            (See `jax.experimental.optimizers <https://jax.readthedocs.io/en/st
            able/jax.experimental.optimizers.html>`_)
        key_or_state : int(2) or :obj:state
            Either a stateless random number generator or the state object of
            an preinitialised optimiser
        main : list of tf.data.Dataset().as_numpy_iterators()
            The simulations generated at the fiducial model parameter values
            used for calculating the covariance of network outputs and their
            derivatives with respect to the physical model parameters (for
            fitting). These are served ``n_per_device`` at at time as a numpy
            iterator from a TensorFlow dataset.
        remaining : list of tf.data.Dataset().as_numpy_iterators()
            The ``n_s - n_d`` simulations generated at the fiducial model
            parameter values used for calculating the covariance of network
            outputs with a derivative counterpart (for fitting). These are
            served ``n_per_device`` at at time as a numpy iterator from a
            TensorFlow dataset.
        host: jax.device
            The main device where the Fisher calculation is performed
        devices: list
            A list of the available jax devices (from ``jax.devices()``)
        n_per_device: int
            Number of simulations to handle at once, this should be as large as
            possible without letting the memory overflow for the best
            performance
        validation_main : list of tf.data.Dataset().as_numpy_iterators()
            The simulations generated at the fiducial model parameter values
            used for calculating the covariance of network outputs and their
            derivatives with respect to the physical model parameters
            (for validation). These are served ``n_per_device`` at at time as a
            numpy iterator from a TensorFlow dataset.
        validation_remaining : list of tf.data.Dataset().as_numpy_iterators()
            The ``n_s - n_d`` simulations generated at the fiducial model
            parameter values used for calculating the covariance of network
            outputs with a derivative counterpart (for validation). Served
            ``n_per_device`` at time as a numpy iterator from a TensorFlow
            dataset.
        """
        super().__init__(
            n_s=n_s,
            n_d=n_d,
            n_params=n_params,
            n_summaries=n_summaries,
            input_shape=input_shape,
            θ_fid=θ_fid,
            model=model,
            key_or_state=key_or_state,
            optimiser=optimiser,
            fiducial=None,
            derivative=None,
            validation_fiducial=None,
            validation_derivative=None,
            host=host,
            devices=devices,
            n_per_device=n_per_device)
        self._set_prebuilt_dataset(
            main, remaining, validation_main, validation_remaining)

    def _set_data(self, fiducial, derivative, validation_fiducial,
                  validation_derivative):
        """Overwritten function to prevent setting fiducial attributes

        Parameters
        ----------
        fiducial : None
        derivative : None
        validation_fiducial : None
        validation_derivative : None
        """
        pass

    def _set_dataset(self, prefetch=None, cache=None):
        """Overwritten function to prevent building dataset, does list check

        Parameters
        ----------
        prefetch : None
        cache : None
        """
        pass

    def _set_prebuilt_dataset(
            self, main, remaining, validation_main, validation_remaining):
        """Set preconstructed dataset iterators

        Parameters
        ----------
        main : list of tf.data.Dataset().as_numpy_iterators()
            The simulations generated at the fiducial model parameter values
            used for calculating the covariance of network outputs and their
            derivatives with respect to the physical model parameters (for
            fitting). These are served ``n_per_device`` at at time as a numpy
            iterator from a TensorFlow dataset.
        remaining : list of tf.data.Dataset().as_numpy_iterators()
            The ``n_s - n_d`` simulations generated at the fiducial model
            parameter values used for calculating the covariance of network
            outputs with a derivative counterpart (for fitting). These are
            served ``n_per_device`` at at time as a numpy iterator from a
            TensorFlow dataset.
        validation_main : list of tf.data.Dataset().as_numpy_iterators()
            The simulations generated at the fiducial model parameter values
            used for calculating the covariance of network outputs and their
            derivatives with respect to the physical model parameters
            (for validation). These are served ``n_per_device`` at at time as a
            numpy iterator from a TensorFlow dataset.
        validation_remaining : list of tf.data.Dataset().as_numpy_iterators()
            The ``n_s - n_d`` simulations generated at the fiducial model
            parameter values used for calculating the covariance of network
            outputs with a derivative counterpart (for validation). Served
            ``n_per_device`` at time as a numpy iterator from a TensorFlow
            dataset.

        Raises
        ------
        ValueError
            if main or remaining are None
        ValueError
            if length of any input list is not equal to number of devices
        TypeError
            if any input is not a list
        """
        self.main = _check_type(main, list, "main", shape=self.n_devices)
        self.remaining = _check_type(
            remaining, list, "remaining", shape=self.n_devices)
        if ((validation_main is not None)
                and (validation_remaining is not None)):
            self.validation_main = _check_type(
                validation_main, list, "validation_main", shape=self.n_devices)
            self.validation_remaining = _check_type(
                validation_remaining, list, "validation_remaining",
                shape=self.n_devices)
            self.validate = True



import tensorflow as tf
import numpy as np
from inspect import signature
import os
import sys


class TFRecords():
    """ Module for writing simulations to TFRecord to be used by the IMNN

    Parameters
    __________
    record_size : int
        approximate maximum size of an individual record (in Mb)
    padding : int
        zero padding size for record name numbering
    input_shape : tuple
        shape of a single simulation
    file : str
        filename to save the records to
    """
    def __init__(self, record_size=150., padding=5):
        """Constructor method

        Parameters
        ----------
        record_size : float
            The desired size of the final TFRecords (in Mb)
        padding : int
            zero padding size for record name numbering
        """
        self.record_size = record_size * int(1e6)
        self.padding = padding
        self.input_shape = None
        self.file = None

    def write_record(
            self, n_sims, get_simulation, fiducial=True, n_params=None,
            validation=False, directory=None, filename=None, start=0):
        """Write all simulations to set of tfrecords

        Parameters
        __________
        n_sims : int
            number of simulations to be written to record
        get_simulation : func
            function (1 or 3 inputs for fiducial or derivative) which returns a
            single simulation as a numpy array
        fiducial : bool (default True)
            whether the simulations are in the fiducial or derivative format
        n_params : int (opt)
            number of parameters in the simulator model (for the derivative)
        validation : bool (default False)
            tag to automatically prepend `validation` to the filename
        directory : str (opt)
            directory to save records. defaults to current directory
        filename : str (opt)
            filename to save records. defaults to `fiducial` and `derivative`
            depending on the value of `fiducial`
        start : int (opt)
            value to start seed at
        """
        self.check_func(get_simulation, fiducial)
        self.check_params(n_params, fiducial)
        self.file = self.get_file(directory, filename, fiducial, validation)
        serialise = self.get_serialiser(fiducial, get_simulation, n_params)
        record = True
        counter = 0
        simulation = self.get_initial_seed(fiducial, start)
        while record:
            with tf.io.TFRecordWriter(".".join((
                    "_".join((
                        self.file,
                        "{}".format(counter).zfill(self.padding))),
                    "tfrecords"))) as self.writer:
                while self.get_seed(simulation, fiducial) < n_sims:
                    simulation = serialise(simulation, counter)
                    if self.check_size(counter):
                        counter += 1
                        break
            if self.get_seed(simulation, fiducial) == n_sims:
                record = False

    def fiducial_serialiser(self, seed, counter, get_simulation):
        """Serialises a fiducial simulation

        Takes a seed index and a function to get a simulation at that given
        index (and a counter for printing the number of files made). This seed
        and simulation are then serialised and written to file.

        Parameters
        ----------
        seed : int
            index to grab a simulation at
        counter : int
            the number record which is being written to
        get_simulation : fn
            a function which takes an index and returns a simulation
            corresponding to that index

        Returns
        -------
        int:
            the input seed value increased by 1
        """
        print("seed={}, record={}".format(seed, counter), end="\r")
        data = get_simulation(seed)
        if self.input_shape is None:
            self.input_shape = data.shape
        data = data.astype(np.float32)
        data = data.tostring()
        feature = {"data": self._bytes_feature(data),
                   "seed": self._int64_feature(seed)}
        example = tf.train.Example(features=tf.train.Features(feature=feature))
        self.writer.write(example.SerializeToString())
        seed += 1
        return seed

    def derivative_serialiser(
            self, simulation, counter, get_simulation, n_params):
        """Serialises a numerical derivative simulation

        Takes a tuple containing a seed index, an index describing whether the
        corresponding simulation is produced below or above the fiducial
        parameter values and an index describing which parameter the simulation
        is going to be a derivative of. These are used to sequentially get
        simulations from a function until either all of the simulations for the
        derivative of a single seed are collected or until the record is full
        at which it breaks out of the loop to create the next record and
        continue there.

        Parameters
        ----------
        simulation : tuple
            - *(int)* -- index to grab a simulation at
            - *(int)* -- index describing whether generated above or below fid
            - *(int)* -- index for the parameter of interest
        counter : int
            the number record which is being written to
        get_simulation : fn
            a function which takes an index and returns a simulation
            corresponding to that index
        n_params : int
            the number of parameters in the numerical derivative

        Returns
        -------
        tuple
            - *(int)* -- index up to where the simulation was grabbed
            - *(int)* -- index describing whether generated above or below fid
            - *(int)* -- index for the parameter of interest
        """
        seed, derivative, parameter = simulation
        break_out = False
        while derivative < 2:
            while parameter < n_params:
                print("seed={}, derivative={}, parameter={}, record={}".format(
                    seed, derivative, parameter, counter), end="\r")
                data = get_simulation(seed, derivative, parameter)
                data = data.astype(np.float32)
                data = data.tostring()
                feature = {"data": self._bytes_feature(data),
                           "seed": self._int64_feature(seed),
                           "derivative": self._int64_feature(derivative),
                           "parameter": self._int64_feature(parameter)}
                example = tf.train.Example(
                    features=tf.train.Features(feature=feature))
                self.writer.write(example.SerializeToString())
                parameter += 1
                if self.check_size(counter):
                    break_out = True
                    break
                else:
                    break_out = False
            if break_out:
                break
            else:
                derivative += 1
                parameter = 0
        if not break_out:
            seed += 1
            derivative = 0
        return (seed, derivative, parameter)

    def get_serialiser(self, fiducial, get_simulation, n_params):
        """Returns the fiducial or derivative serialiser

        Parameters
        ----------
        fiducial: bool
            whether the fiducial serialiser should be returned or not
        get_simulation: fn
            the function which returns either a simulation or a part of a
            derivative simulation
        n_params: int
            the number of parameters that the derivative is taken wrt

        Returns
        -------
        fn:
            either the fiducial serialiser or the derivative serialiser
        """
        if fiducial:
            return lambda inds, counter: self.fiducial_serialiser(
                inds, counter, get_simulation)
        else:
            return lambda inds, counter: self.derivative_serialiser(
                inds, counter, get_simulation, n_params)

    def parser(self, example):
        """Maps a serialised example simulation into its float32 representation

        Parameters
        ----------
        example: str
            The serialised string to be parsed

        Returns
        -------
        float(input_shape):
            The parsed numerical form of the serialised input

        Todo
        ----
        Will fail if input_shape is not set, with no checking. Something should
        be done about this
        """
        parsed_example = tf.io.parse_single_example(
            example, {"data": tf.io.FixedLenFeature([], tf.string)})
        return tf.reshape(
            tf.io.decode_raw(parsed_example["data"], tf.float32),
            self.input_shape)

    def derivative_parser(self, example, n_params=None):
        """Maps a serialised example derivative into its float32 representation

        This is the parser for an analytical or automatic derivative of the
        simulation with respect to model parameters (should be serialised with
        ``fiducial_serialiser``).

        Parameters
        ----------
        example: str
            The serialised string to be parsed
        n_params: int or None, default=None
            The number of parameters in the derivative. This is required but
            named here for use with functools.partial

        Returns
        -------
        float(input_shape, n_params):
            The parsed numerical form of the serialised input

        Todo
        ----
        Will fail if input_shape is not set, with no checking. Something should
        be done about this
        """
        parsed_example = tf.io.parse_single_example(
            example, {"data": tf.io.FixedLenFeature([], tf.string)})
        return tf.reshape(
            tf.io.decode_raw(parsed_example["data"], tf.float32),
            self.input_shape + (n_params,))

    def numerical_derivative_parser(self, example, n_params=None):
        """Maps a serialised example derivative into its float32 representation

        This is the parser for all the simulations necessary for making a
        numerical derivative of a simulation with respect to model parameters
        (should be serialised with ``derivative_serialiser``).

        Parameters
        ----------
        example: str
            The serialised string to be parsed
        n_params: int or None, default=None
            The number of parameters in the derivative. This is required but
            named here for use with functools.partial

        Returns
        -------
        float(2, n_params, input_shape):
            The parsed numerical form of the serialised input

        Todo
        ----
        Will fail if input_shape is not set, with no checking. Something should
        be done about this
        """
        parsed_example = tf.io.parse_single_example(
            example, {"data": tf.io.FixedLenFeature([], tf.string)})
        return tf.reshape(
            tf.io.decode_raw(parsed_example["data"], tf.float32),
            (2, n_params) + self.input_shape)

    def get_file(self, directory, filename, fiducial, validation):
        """Constructs filepath and name that records will be saved to

        Parameters
        ----------
        directory: str
            The full path to where the files should be saved
        filename: str or None
            a filename to save the records to, if None default names are given
            depending on the value of ``fiducial`` and ``validation``
        fiducial: bool
            whether to call a file ``fiducial`` (if ``filename`` is None) or
            ``derivative``
        validation: bool
            whether to prepend ``validation_`` to the filename (if ``filename``
            is None)

        Returns
        -------
        str:
            the filename to save the record to
        """
        if filename is None:
            if fiducial:
                filename = "fiducial"
            else:
                filename = "derivative"
            if validation:
                filename = "validation_{}".format(filename)
        if directory is not None:
            file = "/".join((directory, filename))
        else:
            file = filename
        return file

    def check_size(self, counter):
        """Checks the size of the current record in Mb to see whether its full

        Parameters
        ----------
        counter: int
            The current open record being written to

        Returns
        -------
        bool:
            True if the record has reached (exceeded) the preassigned record
            size
        """
        return os.path.getsize(
            ".".join((
                "_".join((
                    self.file,
                    "{}".format(counter).zfill(self.padding))),
                "tfrecords"))) > self.record_size

    def get_initial_seed(self, fiducial, start):
        """Sets the initial seed index (or set of seeds) to get simulations at

        When constructing a record with a fiducial (or exact derivative) only
        a seed index for the simulation is needed whilst a derivative index and
        a parameter index are also needed if a numerical derivative simulation
        is being collected. Seed indexs will increase incrementally by 1.

        Parameters
        ----------
        fiducial: bool
            if a fiducial simulation record is being constructed
        start: int
            the initial seed index to collect the simulation at

        Returns
        -------
        int or tuple:
            - *(int)* -- the initial seed index to collect the simulations at
            - *(tuple)*
                - *(int)* -- the initial seed index to collect the simulation
                - *(int)* -- whether the simulation is generated below or
                  above the fiducial parameter values
                - *(int)* -- which respect to which parameter the simulation is
                  used to calculate the numerical gradient
        """
        if fiducial:
            return start
        else:
            return (start, 0, 0)

    def get_seed(self, simulation, fiducial):
        """Gets the seed index depending on input

        With a fiducial (or exact derivative) only a seed index for the
        simulation is needed whilst a derivative index and a parameter index
        are also needed if a numerical derivative simulation is being collected

        Parameters
        ----------
        simulation: tuple
            - *(int)* -- the initial seed index to collect the simulation
            - *(int)* -- whether the simulation is generated below or above the
              fiducial parameter values
            - *(int)* -- which respect to which parameter the simulation is
              used to calculate the numerical gradient
        fiducial: bool
            if a fiducial simulation record is being constructed

        Returns
        -------
        int
            the seed index to collect the simulation at
        """
        if fiducial:
            return simulation
        else:
            return simulation[0]

    def check_func(self, get_simulation, fiducial):
        """Checks the simulation grabber takes the correct number of arguments

        Parameters
        ----------
        get_simulation : fn
            a function to get a single simulation
        fiducial :
            if a fiducial simulation record is being constructed

        Todo
        ----
        Should really make these raise ValueErrors rather than using sys to
        shut down
        """
        if fiducial:
            if len(signature(get_simulation).parameters) != 1:
                print("`get_simulations` must be a function which takes a " +
                      "seed only.")
                sys.exit()
        else:
            if len(signature(get_simulation).parameters) != 3:
                print("`get_simulations` must be a function which takes a " +
                      "seed, derivative and parameter as an argument.")
                sys.exit()

    def check_params(self, n_params, fiducial):
        """Checks that n_params is actually given

        Parameters
        ----------
        n_params : int
            the number of parameters in the derivative
        fiducial :
            if a fiducial simulation record is being constructed
        """
        if not fiducial:
            if n_params is None:
                print("`n_params` must be supplied when making derivative " +
                      "record.")
                sys.exit()

    def _bytes_feature(self, value):
        """Makes a serialised byte list from value (converts tensors to numpy)

        Parameters
        ----------
        value : float (possibly Tensor)
            the simulation to be serialised

        Returns
        -------
        byte_list:
            the simulation as a string of bytes
        """
        if isinstance(value, type(tf.constant(0))):
            value = value.numpy()
        return tf.train.Feature(bytes_list=tf.train.BytesList(value=[value]))

    def _int64_feature(self, value):
        """Makes a serialised int list from value

        Parameters
        ----------
        value : int
            the seed to be serialised

        Returns
        -------
        int_list:
            the serialised int list
        """
        return tf.train.Feature(int64_list=tf.train.Int64List(value=[value]))



from imnn.imnn import SimulatorIMNN, AggregatedSimulatorIMNN, GradientIMNN, \
    AggregatedGradientIMNN, DatasetGradientIMNN, NumericalGradientIMNN, \
    AggregatedNumericalGradientIMNN, DatasetNumericalGradientIMNN
import tensorflow as tf


def IMNN(n_s, n_d, n_params, n_summaries, input_shape, θ_fid, model, optimiser,
         key_or_state, simulator=None, fiducial=None, derivative=None,
         main=None, remaining=None, δθ=None, validation_fiducial=None,
         validation_derivative=None, validation_main=None,
         validation_remaining=None, host=None, devices=None, n_per_device=None,
         cache=None, prefetch=None, verbose=True):
    """Selection function to return correct submodule based on inputs

    Because there are many different subclasses to work with specific types of
    simulations (or a simulator) and how their gradients are calculated this
    function provides a way to try and return the desired one based on the data
    passed.

    Parameters
    ----------
    n_s : int
        Number of simulations used to calculate summary covariance
    n_d : int
        Number of simulations used to calculate mean of summary derivative
    n_params : int
        Number of model parameters
    n_summaries : int
        Number of summaries, i.e. outputs of the network
    input_shape : tuple
        The shape of a single input to the network
    θ_fid : float(n_params,)
        The value of the fiducial parameter values used to generate inputs
    model : tuple, len=2
        Tuple containing functions to initialise neural network
        ``fn(rng: int(2), input_shape: tuple) -> tuple, list`` and the
        neural network as a function of network parameters and inputs
        ``fn(w: list, d: float([None], input_shape)) -> float([None],
        n_summaries)``.
        (Essentibly stax-like, see `jax.experimental.stax <https://jax.read
        thedocs.io/en/stable/jax.experimental.stax.html>`_))
    optimiser : tuple, len=3
        Tuple containing functions to generate the optimiser state
        ``fn(x0: list) -> :obj:state``, to update the state from a list of
        gradients ``fn(i: int, g: list, state: :obj:state) -> :obj:state``
        and to extract network parameters from the state
        ``fn(state: :obj:state) -> list``.
        (See `jax.experimental.optimizers <https://jax.readthedocs.io/en/st
        able/jax.experimental.optimizers.html>`_)
    key_or_state : int(2) or :obj:state
        Either a stateless random number generator or the state object of
        an preinitialised optimiser
    simulator : fn, optional requirement
        (:func:`~immn.SimulatorIMNN`, :func:`~immn.AggregatedSimulatorIMNN`)
        A function that generates a single simulation from a random number
        generator and a tuple (or array) of parameter values at which to
        generate the simulations. For the purposes of use in LFI/ABC afterwards
        it is also useful for the simulator to be able to broadcast to a batch
        of simulations on the zeroth axis
        ``fn(int(2,), float([None], n_params)) -> float([None], input_shape)``
    fiducial : float or list, optional requirement
        The simulations generated at the fiducial model parameter values
        used for calculating the covariance of network outputs
        (for fitting)

            - *(float(n_s, input_shape))* -- :func:`~immn.GradientIMNN`,
              :func:`~immn.NumericalGradientIMNN`,
              :func:`~immn.AggregatedGradientIMNN`,
              :func:`~immn.AggregatedNumericalGradientIMNN`
            - *(list of numpy iterators)* --
              :func:`~immn.DatasetNumericalGradientIMNN`

    derivative : float or list, optional requirement
        The simulations generated at parameter values perturbed from the
        fiducial used to calculate the numerical derivative of network
        outputs with respect to model parameters (for fitting)

            - *(float(n_d, input_shape, n_params))* --
              :func:`~immn.GradientIMNN`,
              :func:`~immn.AggregatedGradientIMNN`
            - *(float(n_d, 2, n_params, input_shape))* --
              :func:`~immn.NumericalGradientIMNN`,
              :func:`~immn.AggregatedNumericalGradientIMNN`
            - *(list of numpy iterators)* --
              :func:`~immn.DatasetNumericalGradientIMNN`

    main : list of numpy iterators, optional requirement
        (:func:`~immn.DatasetGradientIMNN`) The simulations generated at the
        fiducial model parameter values used for calculating the covariance of
        network outputs and their derivatives with respect to the physical
        model parameters (for fitting). These are served ``n_per_device`` at a
        time as a numpy iterator from a TensorFlow dataset.
    remaining : list of numpy iterators, optional requirement
        (:func:`~immn.DatasetGradientIMNN`) The ``n_s - n_d`` simulations
        generated at the fiducial model parameter values used for calculating
        the covariance ofnetwork outputs with a derivative counterpart (for
        fitting). These are served ``n_per_device`` at a time as a numpy
        iterator from a TensorFlow dataset.
    δθ : float(n_params,), optional requirement
        (:func:`~immn.NumericalGradientIMNN`,
        :func:`~immn.AggregatedNumericalGradientIMNN`,
        :func:`~immn.DatasetNumericalGradientIMNN`) Size of perturbation to
        model parameters for the numerical derivative
    validation_fiducial : float or list, optional requirement
        The simulations generated at the fiducial model parameter values
        used for calculating the covariance of network outputs
        (for validation)

            - *(float(n_s, input_shape))* --
              :func:`~immn.GradientIMNN`,
              :func:`~immn.NumericalGradientIMNN`,
              :func:`~immn.AggregatedGradientIMNN`,
              :func:`~immn.AggregatedNumericalGradientIMNN`
            - *(list of numpy iterators)* --
              :func:`~immn.DatasetNumericalGradientIMNN`

    validation_derivative : float or list, optional requirement
        The simulations generated at parameter values perturbed from the
        fiducial used to calculate the numerical derivative of network
        outputs with respect to model parameters (for validation)

            - *(float(n_d, input_shape, n_params))* --
              :func:`~immn.GradientIMNN`,
              :func:`~immn.AggregatedGradientIMNN`
            - *(float(n_d, 2, n_params, input_shape))* --
              :func:`~immn.NumericalGradientIMNN`,
              :func:`~immn.AggregatedNumericalGradientIMNN`
            - *(list of numpy iterators)* --
              :func:`~immn.DatasetNumericalGradientIMNN`

    validation_main : list of numpy iterators, optional requirement
        (:func:`~immn.DatasetGradientIMNN`) The simulations generated at the
        fiducial model parameter values used for calculating the covariance of
        network outputs and their derivatives with respect to the physical
        model parameters (for validation). These are served ``n_per_device`` at
        a time as a numpy iterator from a TensorFlow dataset.
    validation_remaining : list of numpy iterators, optional requirement
        (:func:`~immn.DatasetGradientIMNN`) The ``n_s - n_d`` simulations
        generated at the fiducial model parameter values used for calculating
        the covariance of network outputs with a derivative counterpart (for
        validation). These are served ``n_per_device`` at a time as a numpy
        iterator from a TensorFlow dataset.
    host: jax.device, optional requirement
        (:func:`~immn.AggregatedSimulatorIMNN`,
        :func:`~immn.AggregatedGradientIMNN`,
        :func:`~immn.AggregatedNumericalGradientIMNN`,
        :func:`~immn.DatasetGradientIMNN`,
        :func:`~immn.DatasetNumericalGradientIMNN`) The main device where the
        Fisher calculation is performed (something like
        ``jax.devices("cpu")[0]``)
    devices: list, optional requirement
        (:func:`~immn.AggregatedSimulatorIMNN`,
        :func:`~immn.AggregatedGradientIMNN`,
        :func:`~immn.AggregatedNumericalGradientIMNN`,
        :func:`~imnn.DatasetGradientIMNN`,
        :func:`~immn.DatasetNumericalGradientIMNN`) A list of the available jax
        devices (from ``jax.devices()``)
    n_per_device: int, optional requirement
        (:func:`~immn.AggregatedSimulatorIMNN`,
        :func:`~immn.AggregatedGradientIMNN`,
        :func:`~immn.AggregatedNumericalGradientIMNN`,
        :func:`~immn.DatasetGradientIMNN`,
        :func:`~immn.DatasetNumericalGradientIMNN`) Number of simulations to
        handle at once, this should be as large as possible without letting the
        memory overflow for the best performance
    prefetch : tf.data.AUTOTUNE or int or None, optional, default=None
        How many simulation to prefetch in the tensorflow dataset (could be
        used in :func:`~immn.AggregatedGradientIMNN` and
        :func:`~immn.AggregatedNumericalGradientIMNN`)
    cache : bool, optional, default=None
        Whether to cache simulations in the tensorflow datasets (could be used
        in :func:`~immn.AggregatedGradientIMNN` and
        :func:`~immn.AggregatedNumericalGradientIMNN`)
    """

    if isinstance(model, tf.keras.models.Model):
        raise ValueError(
            "`model` cannot be an instance of `tf.keras.models.Model` when " +
            "using JAX mode IMNN. This is now implemented in `imnn_tf` " +
            "which is a legacy version of the imnn using the final version " +
            "the TensorFlow implementation (v0.2.0). Please consider " +
            "installing this via pip: `pip install imnn_tf`")

    if isinstance(optimiser, tf.keras.optimizers.Optimizer):
        raise ValueError(
            "`optimiser` cannot be an instance of " +
            "`tf.keras.optimizers.Optimizer` when using JAX mode IMNN. This " +
            "is now implemented in `imnn_tf` which is a legacy version of " +
            "the imnn using the final version the TensorFlow implementation " +
            "(v0.2.0). Please consider installing this via pip: " +
            "`pip install imnn_tf`")

    aggregate = False

    parameters = {
        "n_s": n_s,
        "n_d": n_d,
        "n_params": n_params,
        "n_summaries": n_summaries,
        "input_shape": input_shape,
        "θ_fid": θ_fid,
        "model": model,
        "optimiser": optimiser,
        "key_or_state": key_or_state}

    if host is None:
        if devices is None:
            if n_per_device is None:
                if cache is None:
                    if prefetch is not None:
                        raise ValueError(
                            "`prefetch` is provided although no other " +
                            "aggregation parameters are passed")
                else:
                    if prefetch is None:
                        raise ValueError(
                            "`cache` is provided although no other " +
                            "aggregation parameters are passed")
                    else:
                        raise ValueError(
                            "`cache` and `prefetch` are provided although " +
                            "no other aggregation parameters are passed")
            else:
                raise ValueError(
                    "`n_per_device` is provided, but `host` and `devices` " +
                    "are missing. If aggregating computation please provide " +
                    "something like `host=jax.devices('cpu')[0]` and " +
                    "`devices=jax.devices()`.")
        else:
            if n_per_device is None:
                raise ValueError(
                    "`devices` is provided, but `host` and `n_per_device` " +
                    "are missing. If aggregating computation please " +
                    "provide something like `host=jax.devices('cpu')[0]` " +
                    "and the number of parallel calls on each device with " +
                    "`n_per_device`.")
            else:
                raise ValueError(
                    "`devices` and `n_per_device` are provided, but `host` " +
                    "is missing. If aggregating computation please provide " +
                    "something like `host=jax.devices('cpu')[0]`")
    else:
        if devices is None:
            if n_per_device is None:
                raise ValueError(
                    "`host` is provided, but `devices` and `n_per_device` " +
                    "are missing. If aggregating computation please provide " +
                    "something like `devices=jax.devices()` and the " +
                    "number of parallel calls on each device with " +
                    "`n_per_device`.")
            else:
                raise ValueError(
                    "`host` and `n_per_device` are provided, but `devices` " +
                    "is missing. If aggregating computation please provide " +
                    "something like `devices=jax.devices()`.")
        else:
            if n_per_device is None:
                raise ValueError(
                    "`host` and `devices` are provided, but `n_per_device` " +
                    "is missing. If aggregating computation please provide " +
                    "the number of parallel calls on each device with " +
                    "`n_per_device`.")
            aggregate = True
            parameters["host"] = host
            parameters["devices"] = devices
            parameters["n_per_device"] = n_per_device

    if simulator is None:
        if fiducial is None:
            if derivative is None:
                if main is None:
                    if remaining is None:
                        raise ValueError(
                            "`simulator` or `fiducial` and `derivative` or " +
                            "`main` and `remaining` argments are necessary")
                    else:
                        raise ValueError(
                            "`remaining` simulations provided, but `main` " +
                            "simulations (with simulations and accompanying " +
                            "derivatives with respect to model parameters) " +
                            "missing.")
                else:
                    if remaining is None:
                        raise ValueError(
                            "`main` simulations provided, but `remaining` " +
                            "simulations are missing. Even if `n_s = n_d` " +
                            "`remaining` needs to be supplied (as a list of " +
                            " numpy iterators with no iterations).")
                    else:
                        if not aggregate:
                            raise ValueError(
                                "aggregation parameters (`host`, `devices` " +
                                "and `n_per_device`) are required when " +
                                "passing `main` and `remaining`")
                        if δθ is not None:
                            raise ValueError(
                                "`δθ` should not be passed when providing " +
                                "`main` and `remaining`. `δθ` is used when " +
                                "calculating numerical gradients, whereas " +
                                "`main` should contain simulations and " +
                                "actual gradients")
                        if cache is not None:
                            raise ValueError(
                                "`cache` should not be passed when passing " +
                                "`main` and `remaining` since the cache " +
                                "should already be built into the dataset " +
                                "if desired")
                        if prefetch is not None:
                            raise ValueError(
                                "`prefetch` should not be passed when " +
                                "passing `main` and `remaining` since " +
                                "prefetching should already be built into " +
                                "the dataset if desired")
                        if validation_fiducial is not None:
                            raise ValueError(
                                "`main` and `remaining` simulations are " +
                                "provided, but `validation_fiducial` is " +
                                "passed too. Validation can only be done " +
                                "using `validation_main` and " +
                                "`validation_remaining` when `main` and " +
                                "`remaining` are supplied.")
                        if validation_derivative is not None:
                            raise ValueError(
                                "`main` and `remaining` simulations are " +
                                "provided, but `validation_derivative` is " +
                                "passed too. Validation can only be done " +
                                "using `validation_main` and " +
                                "`validation_remaining` when `main` and " +
                                "`remaining` are supplied.")
                        if validation_main is not None:
                            if validation_remaining is None:
                                raise ValueError(
                                    "`validation_main` is provided, but " +
                                    "`validation_remaining is missing`")
                            else:
                                parameters["validation_main"] = validation_main
                                parameters["validation_remaining"] = \
                                    validation_remaining
                        else:
                            if validation_remaining is not None:
                                raise ValueError(
                                    "`validation_remaining` is provided, " +
                                    "but `validation_main` is missing")
                        parameters["main"] = main
                        parameters["remaining"] = remaining
                        if verbose:
                            print("`main` and `required` provided, using " +
                                  "`DatasetGradientIMNN`")
                        return DatasetGradientIMNN(**parameters)
            else:
                raise ValueError(
                    "`derivative` supplied but `fiducial` is missing")
        else:
            if derivative is None:
                raise ValueError(
                    "`fiducial` supplied but `derivative` is missing")
            elif (isinstance(fiducial, list)
                    and (not isinstance(derivative, list))):
                raise TypeError("`fiducial` is a list but `derivative` is not")
            elif ((not isinstance(fiducial, list))
                    and isinstance(derivative, list)):
                raise TypeError("`derivative` is a list but `fiducial` is not")
            elif (isinstance(fiducial, list) and (δθ is None)):
                raise TypeError(
                    "`fiducial` and `derivative` should not be lists when " +
                    "passing not `δθ`. If numerical derivatives are " +
                    "intended the please provide `δθ`, and if not, but you " +
                    "want to use a list of dataset iterations then please " +
                    "construct `main` and `remaining`.")
            else:
                if main is not None:
                    if remaining is None:
                        raise ValueError(
                            "`fiducial`, `derivative` and `main` are " +
                            "supplied but only either `main` and " +
                            "`remaining` OR `fiducial` and `derivative` " +
                            "should be passed")
                    else:
                        raise ValueError(
                            "`fiducial` and `derivative` AND `main` and " +
                            "`remaining` are supplied but only either " +
                            "`main` and `remaining` OR `fiducial` and " +
                            "`derivative` should be passed")
                else:
                    if remaining is not None:
                        raise ValueError(
                            "`fiducial`, `derivative` and `remaining` are " +
                            "supplied but only either `main` and " +
                            "`remaining` OR `fiducial` and `derivative` " +
                            "should be passed")
                    if validation_main is not None:
                        raise ValueError(
                            "`fiducial` and `derivative` simulations are " +
                            "provided, but `validation_main` is " +
                            "passed too. Validation can only be done " +
                            "using `validation_fiducial` and " +
                            "`validation_derivative` when `fiducial` and " +
                            "`derivative` are supplied.")
                    if validation_remaining is not None:
                        raise ValueError(
                            "`fiducial` and `derivative` simulations are " +
                            "provided, but `validation_remaining` is " +
                            "passed too. Validation can only be done " +
                            "using `validation_fiducial` and " +
                            "`validation_derivative` when `fiducial` and " +
                            "`derivative` are supplied.")
                    if validation_fiducial is not None:
                        if validation_derivative is None:
                            raise ValueError(
                                "`validation_fiducial` is provided, but " +
                                "`validation_derivative` is missing")
                        elif (isinstance(validation_fiducial, list)
                                and (not isinstance(
                                    validation_derivative, list))):
                            raise TypeError(
                                "`validation_fiducial` is a list but " +
                                "`validation_derivative` is not")
                        elif ((not isinstance(validation_fiducial, list))
                                and isinstance(validation_derivative, list)):
                            raise TypeError(
                                "`validation_derivative` is a list but " +
                                "`validation_fiducial` is not")
                        elif (isinstance(fiducial, list)
                                and (not isinstance(
                                    validation_fiducial, list))):
                            raise TypeError(
                                "`fiducial` and `derivative` are lists but " +
                                "`validation_fiducial` and " +
                                "`validation_derivative` are not")
                        elif ((not isinstance(fiducial, list))
                                and isinstance(validation_fiducial, list)):
                            raise TypeError(
                                "`fiducial` and `derivative` are not lists " +
                                " but `validation_fiducial` and " +
                                "`validation_derivative` are")
                        else:
                            parameters["validation_fiducial"] = \
                                validation_fiducial
                            parameters["validation_derivative"] = \
                                validation_derivative
                    else:
                        if validation_derivative is not None:
                            raise ValueError(
                                "`validation_derivative` is provided, but " +
                                "`validation_fiducial` is missing")
                    parameters["fiducial"] = fiducial
                    parameters["derivative"] = derivative
    else:
        if fiducial is not None:
            raise ValueError(
                "`simulator` and `fiducial` provided, please only provide " +
                "one of these. If using `fiducial` ensure that `derivative` " +
                "is also provided.")
        if derivative is not None:
            raise ValueError(
                "`simulator` and `derivative` provided, please only provide " +
                "one of these. If using `derivative` ensure that `fiducial` " +
                "is also provided.")
        if main is not None:
            raise ValueError(
                "`simulator` and `main` provided, please only provide " +
                "one of these. If using `main` ensure that `remaining` " +
                "is also provided.")
        if remaining is not None:
            raise ValueError(
                "`simulator` and `remaining` provided, please only provide " +
                "one of these. If using `remaining` ensure that `main` " +
                "is also provided.")
        if validation_fiducial is not None:
            raise ValueError(
                "`simulator` and `validation_fiducial` provided, please " +
                "only provide one of these. If using `validation_fiducial` " +
                "ensure that `fiducial`, `derivative` and " +
                "`validation_derivative` are also provided.")
        if validation_derivative is not None:
            raise ValueError(
                "`simulator` and `validation_derivative` provided, please " +
                "only provide one of these. If using " +
                "`validation_derivative`ensure that `fiducial`, " +
                "`derivative` and `validation_fiducial` are also provided.")
        if validation_main is not None:
            raise ValueError(
                "`simulator` and `validation_main` provided, please only " +
                "provide one of these. If using `validation_main` ensure " +
                "that `main`, `remaining` and `validation_remaining` are " +
                "also provided.")
        if validation_remaining is not None:
            raise ValueError(
                "`simulator` and `validation_remaining` provided, please " +
                "only provide one of these. If using `validation_remaining` " +
                "ensure that `main`, `remaining` and `validation_main` are " +
                "also provided.")
        if δθ is not None:
            raise ValueError(
                "`δθ` should not be passed when providing `simulator`. `δθ` " +
                "is used when calculating numerical gradients, whereas the " +
                "automatic derivatives are calculated using the simulator")
        if cache is not None:
            raise ValueError(
                "`cache` should not be passed when passing `simulator` " +
                "since TensorFlow datasets are not used during simulation")
        if prefetch is not None:
            raise ValueError(
                "`prefetch` should not be passed when passing `simulator` " +
                "since TensorFlow datasets are not used during simulation")
        parameters["simulator"] = simulator
        if aggregate:
            if verbose:
                print("`simulator` provided, using AggregatedSimulatorIMNN")
            return AggregatedSimulatorIMNN(**parameters)
        else:
            if verbose:
                print("`simulator` provided, using SimulatorIMNN")
            return SimulatorIMNN(**parameters)

    if cache is not None:
        parameters["cache"] = cache
    if prefetch is not None:
        parameters["prefetch"] = prefetch

    if δθ is None:
        try:
            if derivative.shape != (n_d,) + input_shape + (n_params,):
                raise ValueError(
                    "`derivative.shape` must be " +
                    f"{(n_d,) + input_shape + (2,)} when not doing " +
                    f"numerical gradients, but is {derivative.shape}. If " +
                    "numerical gradients are intended, perhaps `δθ` has not " +
                    "been provided.")
        except AttributeError:
            raise TypeError(
                "`derivative` does not have a `shape` attribute suggesting " +
                "that the wrong type is being passed. `derivative` must be " +
                f"a numpy (jax) array but has type {type(derivative)}")
        if validation_derivative is not None:
            try:
                if validation_derivative.shape != \
                        (n_d,) + input_shape + (n_params,):
                    raise ValueError(
                        "`validation_derivative.shape` must be " +
                        f"{(n_d,) + input_shape + (2,)} when not doing " +
                        "numerical gradients but is " +
                        f"{validation_derivative.shape}. If numerical " +
                        "gradients are intended, perhaps `δθ` has not been " +
                        "provided.")
            except AttributeError:
                raise TypeError(
                    "`validation_derivative does not have a `shape` " +
                    "attribute suggesting that the wrong type is being " +
                    "passed. `validation_derivative` must be a numpy (jax) " +
                    f"array but has type {type(validation_derivative)}")
        if aggregate:
            if verbose:
                print("using AggregatedGradientIMNN")
            return AggregatedGradientIMNN(**parameters)
        else:
            if verbose:
                print("using GradientIMNN")
            return GradientIMNN(**parameters)
    else:
        parameters["δθ"] = δθ
        if isinstance(fiducial, list):
            if not aggregate:
                raise TypeError(
                    "`fiducial` and `derivative` are lists suggesting " +
                    "that they contain numpy iterators over datasets to " +
                    "be used with `DatasetNumericalGradientIMNN`, but " +
                    "no aggregation parameters passed")
            if cache is not None:
                raise ValueError(
                    "`cache` should not be passed when passing `fiducial` " +
                    "and `derivative` as lists since the cache should " +
                    "already be built into the dataset if desired")
            if prefetch is not None:
                raise ValueError(
                    "`prefetch` should not be passed when passing " +
                    "`fiducial` and `derivative` as lists since the " +
                    "prefetching should already be built into the dataset " +
                    "if desired")
            if verbose:
                print("using DatasetNumericalGradientIMNN")
            return DatasetNumericalGradientIMNN(**parameters)
        else:
            try:
                if derivative.shape != (n_d, 2, n_params) + input_shape:
                    raise ValueError(
                        "`derivative.shape` for numerical derivatives must " +
                        f"be {(n_d, 2, n_params) + input_shape} but is " +
                        f"{derivative.shape}. This is (n_d, 2, n_params) + " +
                        "input_shape where [:, 0] are the simulations made " +
                        "below the fiducial and [:, 1] are made above the " +
                        "fiducial.  If numerical gradients are not intended " +
                        "perhaps `δθ` has accidentally been provided.")
            except AttributeError:
                raise TypeError(
                    "`derivative` does not have a `shape` attribute " +
                    "suggesting that the wrong type is being passed. " +
                    "`derivative` must be a numpy (jax) array but has type " +
                    f"{type(derivative)}")
            if validation_derivative is not None:
                try:
                    if validation_derivative.shape != \
                            (n_d, 2, n_params) + input_shape:
                        raise ValueError(
                            "`validation_derivative.shape` for numerical " +
                            "derivatives must be " +
                            f"{(n_d, 2, n_params) + input_shape} but is " +
                            f"{validation_derivative.shape}. This is " +
                            "(n_d, 2, n_params) + input_shape where [:, 0] " +
                            "are the simulations made below the fiducial " +
                            "and [:, 1] are made above the fiducial. If " +
                            "numerical gradients are not intended perhaps " +
                            "`δθ` has accidentally been provided.")
                except AttributeError:
                    raise TypeError(
                        "`validation_derivative does not have a `shape` " +
                        "attribute suggesting that the wrong type is being " +
                        "passed. `validation_derivative` must be a numpy " +
                        "(jax) array but has type " +
                        f"{type(validation_derivative)}")
            if aggregate:
                if verbose:
                    print("using AggregatedNumericalGradientIMNN")
                return AggregatedNumericalGradientIMNN(**parameters)
            else:
                if verbose:
                    print("using NumericalGradientIMNN")
                return NumericalGradientIMNN(**parameters)

    print(
        "I would have thought all options should have been exhausted. " +
        "Please check that all parameters are correct - if they definitely " +
        "are then there is an error in the logic of this code, but you " +
        "probably also know what you are doing enough to directly use the " +
        "specific subclass.")



import jax
import jax.numpy as np
from imnn.imnn._imnn import _IMNN
from imnn.utils import value_and_jacrev
from imnn.utils.utils import _check_input


class GradientIMNN(_IMNN):
    """Information maximising neural network fit using known derivatives

    The outline of the fitting procedure is that a set of :math:`i\\in[1, n_s]`
    simulations :math:`{\\bf d}^i` originally generated at fiducial model
    parameter :math:`{\\bf\\theta}^\\rm{fid}`, and their derivatives
    :math:`\\partial{\\bf d}^i/\\partial\\theta_\\alpha` with respect to model
    parameters are used. The fiducial simulations, :math:`{\\bf d}^i`, are
    passed through a network to obtain summaries, :math:`{\\bf x}^i`, and the
    jax automatic derivative of these summaries with respect to the inputs are
    calculated :math:`\\partial{\\bf x}^i\\partial{\\bf d}^j\\delta_{ij}`. The
    chain rule is then used to calculate

    .. math::
        \\frac{\\partial{\\bf x}^i}{\\partial\\theta_\\alpha} =
        \\frac{\\partial{\\bf x}^i}{\\partial{\\bf d}^j}
        \\frac{\\partial{\\bf d}^j}{\\partial\\theta_\\alpha}

    With :math:`{\\bf x}^i` and
    :math:`\\partial{{\\bf x}^i}/\\partial\\theta_\\alpha` the covariance

    .. math::
        C_{ab} = \\frac{1}{n_s-1}\\sum_{i=1}^{n_s}(x^i_a-\\mu^i_a)
        (x^i_b-\\mu^i_b)

    and the derivative of the mean of the network outputs with respect to the
    model parameters

    .. math::
        \\frac{\\partial\\mu_a}{\\partial\\theta_\\alpha} = \\frac{1}{n_d}
        \\sum_{i=1}^{n_d}\\frac{\\partial{x^i_a}}{\\partial\\theta_\\alpha}

    can be calculated and used form the Fisher information matrix

    .. math::
        F_{\\alpha\\beta} = \\frac{\\partial\\mu_a}{\\partial\\theta_\\alpha}
        C^{-1}_{ab}\\frac{\\partial\\mu_b}{\\partial\\theta_\\beta}.

    The loss function is then defined as

    .. math::
        \\Lambda = -\\log|{\\bf F}| + r(\\Lambda_2) \\Lambda_2

    Since any linear rescaling of a sufficient statistic is also a sufficient
    statistic the negative logarithm of the determinant of the Fisher
    information matrix needs to be regularised to fix the scale of the network
    outputs. We choose to fix this scale by constraining the covariance of
    network outputs as

    .. math::
        \\Lambda_2 = ||{\\bf C}-{\\bf I}|| + ||{\\bf C}^{-1}-{\\bf I}||

    Choosing this constraint is that it forces the covariance to be
    approximately parameter independent which justifies choosing the covariance
    independent Gaussian Fisher information as above. To avoid having a dual
    optimisation objective, we use a smooth and dynamic regularisation strength
    which turns off the regularisation to focus on maximising the Fisher
    information when the covariance has set the scale

    .. math::
        r(\\Lambda_2) = \\frac{\\lambda\\Lambda_2}{\\Lambda_2-\\exp
        (-\\alpha\\Lambda_2)}.

    Once the loss function is calculated the automatic gradient is then
    calculated and used to update the network parameters via the optimiser
    function.

    Parameters
    ----------
    fiducial : float(n_s, input_shape)
        The simulations generated at the fiducial model parameter values used
        for calculating the covariance of network outputs (for fitting)
    derivative : float(n_d, input_shape, n_params)
        The derivative of the simulations with respect to the model parameters
        (for fitting)
    validation_fiducial : float(n_s, input_shape) or None
        The simulations generated at the fiducial model parameter values used
        for calculating the covariance of network outputs (for validation)
    validation_derivative : float(n_d, input_shape, n_params) or None
        The derivative of the simulations with respect to the model parameters
        (for validation)
    """
    def __init__(self, n_s, n_d, n_params, n_summaries, input_shape, θ_fid,
                 model, optimiser, key_or_state, fiducial, derivative,
                 validation_fiducial=None, validation_derivative=None):
        """Constructor method

        Initialises all IMNN attributes, constructs neural network and its
        initial parameter values and creates history dictionary. Also fills the
        simulation attributes (and validation if available).

        Parameters
        ----------
        n_s : int
            Number of simulations used to calculate summary covariance
        n_d : int
            Number of simulations used to calculate mean of summary derivative
        n_params : int
            Number of model parameters
        n_summaries : int
            Number of summaries, i.e. outputs of the network
        input_shape : tuple
            The shape of a single input to the network
        θ_fid : float(n_params,)
            The value of the fiducial parameter values used to generate inputs
        model : tuple, len=2
            Tuple containing functions to initialise neural network
            ``fn(rng: int(2), input_shape: tuple) -> tuple, list`` and the
            neural network as a function of network parameters and inputs
            ``fn(w: list, d: float(None, input_shape)) -> float(None, n_summari
            es)``.
            (Essentibly stax-like, see `jax.experimental.stax <https://jax.read
            thedocs.io/en/stable/jax.experimental.stax.html>`_))
        optimiser : tuple, len=3
            Tuple containing functions to generate the optimiser state
            ``fn(x0: list) -> :obj:state``, to update the state from a list of
            gradients ``fn(i: int, g: list, state: :obj:state) -> :obj:state``
            and to extract network parameters from the state
            ``fn(state: :obj:state) -> list``.
            (See `jax.experimental.optimizers <https://jax.readthedocs.io/en/st
            able/jax.experimental.optimizers.html>`_)
        key_or_state : int(2) or :obj:state
            Either a stateless random number generator or the state object of
            an preinitialised optimiser
        fiducial : float(n_s, input_shape)
            The simulations generated at the fiducial model parameter values
            used for calculating the covariance of network outputs
            (for fitting)
        derivative : float(n_d, input_shape, n_params)
            The derivative of the simulations with respect to the model
            parameters (for fitting)
        validation_fiducial : float(n_s, input_shape) or None, default=None
            The simulations generated at the fiducial model parameter values
            used for calculating the covariance of network outputs
            (for validation)
        validation_derivative : float(n_d, input_shape, n_params) or None
            The derivative of the simulations with respect to the model
            parameters (for validation)
        """
        super().__init__(
            n_s=n_s,
            n_d=n_d,
            n_params=n_params,
            n_summaries=n_summaries,
            input_shape=input_shape,
            θ_fid=θ_fid,
            model=model,
            key_or_state=key_or_state,
            optimiser=optimiser)
        self._set_data(fiducial, derivative, validation_fiducial,
                       validation_derivative)

    def _set_data(self, fiducial, derivative, validation_fiducial,
                  validation_derivative):
        """Checks and sets data attributes with the correct shape

        Parameters
        ----------
        fiducial : float(n_s, input_shape)
            The simulations generated at the fiducial model parameter values
            used for calculating the covariance of network outputs
            (for fitting)
        derivative : float(n_d, input_shape, n_params)
            The derivative of the simulations with respect to the model
            parameters (for fitting)
        validation_fiducial : float(n_s, input_shape) or None, default=None
            The simulations generated at the fiducial model parameter values
            used for calculating the covariance of network outputs
            (for validation). Sets ``validate = True`` attribute if provided
        validation_derivative : float(n_d, input_shape, n_params) or None
            The derivative of the simulations with respect to the model
            parameters (for validation). Sets ``validate = True`` attribute if
            provided
        """
        self.fiducial = _check_input(
            fiducial, (self.n_s,) + self.input_shape, "fiducial")
        self.derivative = _check_input(
            derivative, (self.n_d,) + self.input_shape + (self.n_params,),
            "derivative")
        if ((validation_fiducial is not None)
                and (validation_derivative is not None)):
            self.validation_fiducial = _check_input(
                validation_fiducial, (self.n_s,) + self.input_shape,
                "validation_fiducial")
            self.validation_derivative = _check_input(
                validation_derivative,
                (self.n_d,) + self.input_shape + (self.n_params,),
                "validation_derivative")
            self.validate = True

    def get_summaries(self, w, key=None, validate=False):
        """Gets all network outputs and derivatives wrt model parameters

        Selects either the fitting or validation sets and passes them through
        the network to get the network outputs.

        Parameters
        ----------
        w : list or None, default=None
            The network parameters if wanting to calculate the Fisher
            information with a specific set of network parameters
        key : int(2,) or None, default=None
            A random number generator for generating simulations on-the-fly
        validate : bool, default=False
            Whether to get summaries of the validation set

        Returns
        -------
        float(n_s, n_summaries):
            The set of all network outputs used to calculate the covariance
        float(n_d, n_summaries, n_params):
            The derivative of the network ouputs wrt the model parameters

        Methods
        -------
        get_derivatives:
            Calculates the jacobian of the network output and its value
        """
        def get_derivatives(simulation):
            """Calculates the jacobian of the network output and its value

            Parameters
            ----------
            simulation: float(input_shape)
                The fiducial simulation to calculate the Jacobian of

            Returns
            -------
            float(n_summaries):
                The network output (summary) of the simulation
            float(input_shape, n_summaries):
                The derivative of the network outputs wrt the inputs
            """
            x, dx_dd = value_and_jacrev(self.model, argnums=1)(w, simulation)
            return x, dx_dd.T

        if validate:
            fiducial = self.validation_fiducial
            derivative = self.validation_derivative
        else:
            fiducial = self.fiducial
            derivative = self.derivative
        if self.n_d < self.n_s:
            summaries, dx_dd = jax.vmap(get_derivatives)(fiducial[:self.n_d])
            summaries = np.vstack([
                summaries,
                self.model(w, fiducial[self.n_d:])])
        else:
            summaries, dx_dd = jax.vmap(get_derivatives)(fiducial)
        return summaries, np.einsum("i...j,i...k->ijk", dx_dd, derivative)
